/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.internal.ClientComms;
import org.eclipse.paho.client.mqttv3.internal.ConnectActionListener;
import org.eclipse.paho.client.mqttv3.internal.ExceptionHelper;
import org.eclipse.paho.client.mqttv3.internal.LocalNetworkModule;
import org.eclipse.paho.client.mqttv3.internal.NetworkModule;
import org.eclipse.paho.client.mqttv3.internal.SSLNetworkModule;
import org.eclipse.paho.client.mqttv3.internal.TCPNetworkModule;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttDisconnect;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPublish;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttSubscribe;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttUnsubscribe;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;
import org.eclipse.paho.client.mqttv3.persist.MqttDefaultFilePersistence;

import java.text.MessageFormat;
import java.util.Hashtable;
import java.util.Properties;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocketFactory;

/**
 * Lightweight client for talking to an MQTT server using non-blocking methods
 * that allow an operation to run in the background.
 * <p/>
 * <p>
 * This class implements the non-blocking {@link IMqttAsyncClient} client interface allowing applications to initiate MQTT actions and then carry on working while the MQTT action completes on a
 * background thread. This implementation is compatible with all Java SE runtimes from 1.4.2 and up.
 * </p>
 * <p>
 * An application can connect to an MQTT server using:
 * <ul>
 * <li>A plain TCP socket
 * <li>A secure SSL/TLS socket
 * </ul>
 * </p>
 * <p>
 * To enable messages to be delivered even across network and client restarts messages need to be safely stored until the message has been delivered at the requested quality of service. A pluggable
 * persistence mechanism is provided to store the messages.
 * </p>
 * <p>
 * By default {@link MqttDefaultFilePersistence} is used to store messages to a file. If persistence is set to null then messages are stored in memory and hence can be lost if the client, Java runtime
 * or device shuts down.
 * </p>
 * <p>
 * If connecting with {@link MqttConnectOptions#setCleanSession(boolean)} set to true it is safe to use memory persistence as all state is cleared when a client disconnects. If connecting with
 * cleanSession set to false in order to provide reliable message delivery then a persistent message store such as the default one should be used.
 * </p>
 * <p>
 * The message store interface is pluggable. Different stores can be used by implementing the {@link MqttClientPersistence} interface and passing it to the clients constructor.
 * </p>
 *
 * @see IMqttAsyncClient
 */
public class MqttAsyncClient implements IMqttAsyncClient { // DestinationProvider {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "MqttAsyncClient";

    private static final String CLIENT_ID_PREFIX = "paho-";
    private static final long QUIESCE_TIMEOUT = 30000; // ms
    private static final long DISCONNECT_TIMEOUT = 10000; // ms
    private final String mClientId;
    private final String mServerURI;
    private final ClientComms mComms;
    private final Hashtable<String, MqttTopic> mTopicMap;
    private MqttClientPersistence mPersistence;

    /**
     * Create an MqttAsyncClient that is used to communicate with an MQTT server.
     * <p>
     * The address of a server can be specified on the constructor.
     * setServerURIs} method on MqttConnectOptions.
     * <p/>
     * <p>
     * The <code>serverURI</code> parameter is typically used with the the <code>clientId</code> parameter to form a key. The key is used to store and reference messages while they are being
     * delivered. Hence the serverURI specified on the constructor must still be specified even if a list of servers is specified on an MqttConnectOptions object. The serverURI on the constructor must
     * remain the same across restarts of the client for delivery of messages to be maintained from a given client to a given server or set of servers.
     * <p/>
     * <p>
     * The address of the server to connect to is specified as a URI. Two types of connection are supported <code>tcp://</code> for a TCP connection and <code>ssl://</code> for a TCP connection
     * secured by SSL/TLS. For example:
     * <ul>
     * <li><code>tcp://localhost:1883</code></li>
     * <li><code>ssl://localhost:8883</code></li>
     * </ul>
     * If the port is not specified, it will default to 1883 for <code>tcp://</code>" URIs, and 8883 for <code>ssl://</code> URIs.
     * </p>
     * <p/>
     * <p>
     * A client identifier <code>clientId</code> must be specified and be less that 65535 characters. It must be unique across all clients connecting to the same server. The clientId is used by the
     * server to store data related to the client, hence it is important that the clientId remain the same when connecting to a server if durable subscriptions or reliable messaging are required.
     * <p>
     * A convenience method is provided to generate a random client id that should satisfy this criteria - {@link #generateClientId()}. As the client identifier is used by the server to identify a
     * client when it reconnects, the client must use the same identifier between connections if durable subscriptions or reliable delivery of messages is required.
     * </p>
     * <p>
     * In Java SE, SSL can be configured in one of several ways, which the client will use in the following order:
     * </p>
     * <ul>
     * <li><strong>Supplying an <code>SSLSocketFactory</code></strong> - applications can use {@link MqttConnectOptions#setSocketFactory(ProdGroupCallSSLSocketFactory)} to supply a factory with the appropriate SSL
     * settings.</li>
     * <li><strong>SSL Properties</strong> - applications can supply SSL settings as a simple Java Properties using {@link MqttConnectOptions#setSSLProperties(Properties)}.</li>
     * <li><strong>Use JVM settings</strong> - There are a number of standard Java system properties that can be used to configure key and trust stores.</li>
     * </ul>
     * <p/>
     * <p>
     * In Java ME, the platform settings are used for SSL connections.
     * </p>
     * <p/>
     * <p/>
     * An instance of the default persistence mechanism {@link MqttDefaultFilePersistence} is used by the client. To specify a different persistence mechanism or to turn off mPersistence, use the
     * {@link #MqttAsyncClient(String, String, MqttClientPersistence)} constructor.
     *
     * @param serverURI the address of the server to connect to, specified as a URI.
     * @param clientId  a client identifier that is unique on the server being connected to
     * @throws IllegalArgumentException if the URI does not start with
     *                                  "tcp://", "ssl://" or "local://".
     * @throws IllegalArgumentException if the clientId is null or is greater than 65535 characters in length
     * @throws MqttException            if any other problem was encountered
     */
    public MqttAsyncClient(String serverURI, String clientId) throws MqttException {
        this(serverURI, clientId, new MqttDefaultFilePersistence());
    }

    public MqttAsyncClient(String serverURI, String clientId, MqttClientPersistence persistence) throws MqttException {
        this(serverURI, clientId, persistence, new TimerPingSender());
    }

    /**
     * Create an MqttAsyncClient that is used to communicate with an MQTT server.
     * <p>
     * The address of a server can be specified on the constructor.
     * setServerURIs} method on MqttConnectOptions.
     * <p/>
     * <p>
     * The <code>serverURI</code> parameter is typically used with the the <code>clientId</code> parameter to form a key. The key is used to store and reference messages while they are being
     * delivered. Hence the serverURI specified on the constructor must still be specified even if a list of servers is specified on an MqttConnectOptions object. The serverURI on the constructor must
     * remain the same across restarts of the client for delivery of messages to be maintained from a given client to a given server or set of servers.
     * <p/>
     * <p>
     * The address of the server to connect to is specified as a URI. Two types of connection are supported <code>tcp://</code> for a TCP connection and <code>ssl://</code> for a TCP connection
     * secured by SSL/TLS. For example:
     * <ul>
     * <li><code>tcp://localhost:1883</code></li>
     * <li><code>ssl://localhost:8883</code></li>
     * </ul>
     * If the port is not specified, it will default to 1883 for <code>tcp://</code>" URIs, and 8883 for <code>ssl://</code> URIs.
     * </p>
     * <p/>
     * <p>
     * A client identifier <code>clientId</code> must be specified and be less that 65535 characters. It must be unique across all clients connecting to the same server. The clientId is used by the
     * server to store data related to the client, hence it is important that the clientId remain the same when connecting to a server if durable subscriptions or reliable messaging are required.
     * <p>
     * A convenience method is provided to generate a random client id that should satisfy this criteria - {@link #generateClientId()}. As the client identifier is used by the server to identify a
     * client when it reconnects, the client must use the same identifier between connections if durable subscriptions or reliable delivery of messages is required.
     * </p>
     * <p>
     * In Java SE, SSL can be configured in one of several ways, which the client will use in the following order:
     * </p>
     * <ul>
     * <li><strong>Supplying an <code>SSLSocketFactory</code></strong> - applications can use {@link MqttConnectOptions#setSocketFactory(ProdGroupCallSSLSocketFactory)} to supply a factory with the appropriate SSL
     * settings.</li>
     * <li><strong>SSL Properties</strong> - applications can supply SSL settings as a simple Java Properties using {@link MqttConnectOptions#setSSLProperties(Properties)}.</li>
     * <li><strong>Use JVM settings</strong> - There are a number of standard Java system properties that can be used to configure key and trust stores.</li>
     * </ul>
     * <p/>
     * <p>
     * In Java ME, the platform settings are used for SSL connections.
     * </p>
     * <p>
     * A persistence mechanism is used to enable reliable messaging. For messages sent at qualities of service (QoS) 1 or 2 to be reliably delivered, messages must be stored (on both the client and
     * server) until the delivery of the message is complete. If messages are not safely stored when being delivered then a failure in the client or server can result in lost messages. A pluggable
     * persistence mechanism is supported via the {@link MqttClientPersistence} interface. An implementer of this interface that safely stores messages must be specified in order for delivery of
     * messages to be reliable. In addition {@link MqttConnectOptions#setCleanSession(boolean)} must be set to false. In the event that only QoS 0 messages are sent or received or cleanSession is set
     * to true then a safe store is not needed.
     * </p>
     * <p>
     * An implementation of file-based persistence is provided in class {@link MqttDefaultFilePersistence} which will work in all Java SE based systems. If no persistence is needed, the persistence
     * parameter can be explicitly set to <code>null</code>.
     * </p>
     *
     * @param serverURI   the address of the server to connect to, specified as a URI.
     * @param clientId    a client identifier that is unique on the server being connected to
     * @param persistence the persistence class to use to store in-flight message. If null then the
     *                    default persistence mechanism is used
     * @throws IllegalArgumentException if the URI does not start with
     *                                  "tcp://", "ssl://" or "local://"
     * @throws IllegalArgumentException if the clientId is null or is greater than 65535 characters in length
     * @throws MqttException            if any other problem was encountered
     */
    public MqttAsyncClient(String serverURI, String clientId, MqttClientPersistence persistence, MqttPingSender pingSender) throws MqttException, IllegalArgumentException {
        if (clientId == null) { //Support empty client Id, 3.1.1 standard
            throw new IllegalArgumentException("Null clientId");
        }
        // Count characters, surrogate pairs count as one character.
        int clientIdLength = 0;
        for (int i = 0; i < clientId.length() - 1; i++) {
            if (Character_isHighSurrogate(clientId.charAt(i))) {
                i++;
            }
            clientIdLength++;
        }
        if (clientIdLength > 65535) {
            throw new IllegalArgumentException("ClientId longer than 65535 characters");
        }

        MqttConnectOptions.validateURI(serverURI);

        mServerURI = serverURI;
        mClientId = clientId;

        mPersistence = persistence;
        if (mPersistence == null) {
            mPersistence = new MemoryPersistence();
        }

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[MqttAsyncClient] " + MessageFormat.format("<init> ClientID={0} ServerURI={1} PersistenceType={2}",
                    new Object[]{clientId, serverURI, persistence}));
        }

        mPersistence.open(clientId, serverURI);
        mComms = new ClientComms(this, this.mPersistence, pingSender);
        mPersistence.close();
        mTopicMap = new Hashtable<String, MqttTopic>();

    }

    /**
     * @param ch
     * @return returns 'true' if the character is a high-surrogate code unit
     */
    private static boolean Character_isHighSurrogate(char ch) {
        char MIN_HIGH_SURROGATE = '\uD800';
        char MAX_HIGH_SURROGATE = '\uDBFF';
        return ch >= MIN_HIGH_SURROGATE && ch <= MAX_HIGH_SURROGATE;
    }

    /**
     * Factory method to create the correct network module, based on the
     * supplied address URI.
     * 서버 타입에 맞는 NetworkModule 구현체를 생성한다.
     *
     * @param address the URI for the server.
     * @param options Connect options
     * @return a network module appropriate to the specified address.
     */
    private NetworkModule createNetworkModule(String address, MqttConnectOptions options) throws MqttException, MqttSecurityException, IllegalArgumentException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[createNetworkModule] " + MessageFormat.format("URI={0}", new Object[]{address}));
        }

        NetworkModule netModule;
        String shortAddress;
        String host;
        int port;
        SocketFactory factory = options.getSocketFactory();

        int serverURIType = MqttConnectOptions.validateURI(address);

        switch (serverURIType) {
            case MqttConnectOptions.URI_TYPE_TCP:
                shortAddress = address.substring(6);
                host = getHostName(shortAddress);
                port = getPort(shortAddress, 1883);
                if (factory == null) {
                    factory = SocketFactory.getDefault();
                } else if (factory instanceof SSLSocketFactory) {
                    throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_SOCKET_FACTORY_MISMATCH);
                }
                netModule = new TCPNetworkModule(factory, host, port, mClientId, options.getImsNetwork());
                ((TCPNetworkModule) netModule).setConnectTimeout(options.getConnectionTimeout());
                break;
            case MqttConnectOptions.URI_TYPE_SSL:
                shortAddress = address.substring(6);
                host = getHostName(shortAddress);
                port = getPort(shortAddress, 8883);
                /*
                 * SSLSocketFactoryFactory factoryFactory = null;
                 * if (factory == null) {
                 * factoryFactory = new SSLSocketFactoryFactory();
                 * Properties sslClientProps = options.getSSLProperties();
                 * if (null != sslClientProps)
                 * factoryFactory.initialize(sslClientProps, null);
                 * factory = factoryFactory.createSocketFactory(null);
                 * }
                 * else
                 */
                if (factory instanceof SSLSocketFactory == false) {
                    throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_SOCKET_FACTORY_MISMATCH);
                }

                // Create the network module...
                netModule = new SSLNetworkModule(factory, host, port, mClientId, options.getImsNetwork());
                ((SSLNetworkModule) netModule).setSSLhandshakeTimeout(options.getConnectionTimeout());
                // Ciphers suites need to be set, if they are available
                /*
                 * if (factoryFactory != null) {
                 * String[] enabledCiphers = factoryFactory.getEnabledCipherSuites(null);
                 * if (enabledCiphers != null) {
                 * ((SSLNetworkModule) netModule).setEnabledCiphers(enabledCiphers);
                 * }
                 * }
                 */
                break;
            case MqttConnectOptions.URI_TYPE_LOCAL:
                netModule = new LocalNetworkModule(address.substring(8));
                break;
            default:
                // This shouldn't happen, as long as validateURI() has been called.
                netModule = null;
        }
        return netModule;
    }

    private int getPort(String uri, int defaultPort) {
        int port;
        int portIndex = uri.lastIndexOf(':');
        if (portIndex == -1) {
            port = defaultPort;
        } else {
            port = Integer.parseInt(uri.substring(portIndex + 1));
        }
        return port;
    }

    private String getHostName(String uri) {
        int schemeIndex = uri.lastIndexOf('/');
        int portIndex = uri.lastIndexOf(':');
        if (portIndex == -1) {
            portIndex = uri.length();
        }
        return uri.substring(schemeIndex + 1, portIndex);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#connect(java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken connect(Object userContext, IMqttActionListener callback)
            throws MqttException, MqttSecurityException, MqttPersistenceException, IllegalArgumentException {
        return this.connect(new MqttConnectOptions(), userContext, callback);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#connect()
     */
    @Override
    public IMqttToken connect() throws MqttException, MqttSecurityException, MqttPersistenceException, IllegalArgumentException {
        return this.connect(null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#connect(MqttConnectOptions)
     */
    @Override
    public IMqttToken connect(MqttConnectOptions options) throws MqttException, MqttSecurityException, MqttPersistenceException, IllegalArgumentException {
        return this.connect(options, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#connect(MqttConnectOptions, java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken connect(MqttConnectOptions options, Object userContext, IMqttActionListener callback)
            throws MqttException, MqttSecurityException, MqttPersistenceException, IllegalArgumentException {
        if (mComms.isConnected()) {
            throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_CONNECTED);
        }
        if (mComms.isConnecting()) {
            throw new MqttException(MqttException.REASON_CODE_CONNECT_IN_PROGRESS);
        }
        if (mComms.isDisconnecting()) {
            throw new MqttException(MqttException.REASON_CODE_CLIENT_DISCONNECTING);
        }
        if (mComms.isClosed()) {
            throw new MqttException(MqttException.REASON_CODE_CLIENT_CLOSED);
        }

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[connect] " + MessageFormat.format("cleanSession={0} connectionTimeout={1} TimekeepAlive={2} userName={3} password={4} will={5} userContext={6} callback={7}",
                    new Object[]{
                            Boolean.valueOf(options.isCleanSession()),
                            Integer.valueOf(options.getConnectionTimeout()),
                            Integer.valueOf(options.getKeepAliveInterval()),
                            options.getUserName(),
                            null == options.getPassword() ? "[null]" : "[notnull]",
                            null == options.getWillMessage() ? "[null]" : "[notnull]",
                            userContext,
                            callback}));
        }

        // NetworkModule 생성.
        mComms.setNetworkModule(createNetworkModule(mServerURI, options));

        // 토큰 생성.
        // Insert our own callback to iterate through the URIs till the connect succeeds
        MqttToken userToken = new MqttToken();

        // 토큰에 리스너 등록.
        ConnectActionListener connectActionListener = new ConnectActionListener(this, mPersistence, mComms, options, userToken, userContext, callback);
        userToken.setActionCallback(connectActionListener);
        userToken.setUserContext(this);

        connectActionListener.connect();

        return userToken;
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#disconnect(java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken disconnect(Object userContext, IMqttActionListener callback) throws MqttException {
        return disconnect(QUIESCE_TIMEOUT, userContext, callback);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#disconnect()
     */
    @Override
    public IMqttToken disconnect() throws MqttException {
        return disconnect(null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#disconnect(long)
     */
    @Override
    public IMqttToken disconnect(long quiesceTimeout) throws MqttException {
        return disconnect(quiesceTimeout, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#disconnect(long, java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken disconnect(long quiesceTimeout, Object userContext, IMqttActionListener callback) throws MqttException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[disconnect] " + MessageFormat.format("> quiesceTimeout={0} userContext={1} callback={2}",
                    new Object[]{Long.valueOf(quiesceTimeout), userContext, callback}));
        }

        MqttToken token = new MqttToken();
        token.setActionCallback(callback);
        token.setUserContext(userContext);

        MqttDisconnect disconnect = new MqttDisconnect();
        try {
            mComms.disconnect(disconnect, quiesceTimeout, token);
        } catch (MqttException ex) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[disconnect] " + "< exception");
            }
            throw ex;
        }
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[disconnect] " + "<");
        }

        return token;
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#disconnectForcibly()
     */
    @Override
    public void disconnectForcibly() throws MqttException {
        disconnectForcibly(QUIESCE_TIMEOUT, DISCONNECT_TIMEOUT);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#disconnectForcibly(long)
     */
    @Override
    public void disconnectForcibly(long disconnectTimeout) throws MqttException {
        disconnectForcibly(QUIESCE_TIMEOUT, disconnectTimeout);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#disconnectForcibly(long, long)
     */
    @Override
    public void disconnectForcibly(long quiesceTimeout, long disconnectTimeout) throws MqttException {
        mComms.disconnectForcibly(quiesceTimeout, disconnectTimeout);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#isConnected()
     */
    @Override
    public boolean isConnected() {
        return mComms.isConnected();
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#getClientId()
     */
    @Override
    public String getClientId() {
        return mClientId;
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#getServerURI()
     */
    @Override
    public String getServerURI() {
        return mServerURI;
    }

    /**
     * Get a topic object which can be used to publish messages.
     * <p>
     * There are two alternative methods that should be used in preference to this one when publishing a message:
     * <ul>
     * <li>{@link MqttAsyncClient#publish(String, MqttMessage)} to publish a message in a non-blocking manner or
     * <li>{@link MqttClient#publish(String, MqttMessage)} to publish a message in a blocking manner
     * </ul>
     * </p>
     * <p>
     * When you build an application, the design of the topic tree should take into account the following principles of topic name syntax and semantics:
     * </p>
     * <p/>
     * <ul>
     * <li>A topic must be at least one character long.</li>
     * <li>Topic names are case sensitive. For example, <em>ACCOUNTS</em> and <em>Accounts</em> are two different topics.</li>
     * <li>Topic names can include the space character. For example, <em>Accounts
     * payable</em> is a valid topic.</li>
     * <li>A leading "/" creates a distinct topic. For example, <em>/finance</em> is different from <em>finance</em>. <em>/finance</em> matches "+/+" and "/+", but not "+".</li>
     * <li>Do not include the null character (Unicode<samp class="codeph"> \x0000</samp>) in any topic.</li>
     * </ul>
     * <p/>
     * <p>
     * The following principles apply to the construction and content of a topic tree:
     * </p>
     * <p/>
     * <ul>
     * <li>The length is limited to 64k but within that there are no limits to the number of levels in a topic tree.</li>
     * <li>There can be any number of root nodes; that is, there can be any number of topic trees.</li>
     * </ul>
     * </p>
     *
     * @param topic the topic to use, for example "finance/stock/ibm".
     * @return an MqttTopic object, which can be used to publish messages to
     * the topic.
     * String topic에 해당하는 MqttTopic 생성. 있으면 있는대로.
     * @throws IllegalArgumentException if the topic contains a '+' or '#'
     *                                  wildcard character.
     */
    protected MqttTopic getTopic(String topic) throws IllegalArgumentException {
        MqttTopic.validate(topic, false/* wildcards NOT allowed */);

        MqttTopic result = mTopicMap.get(topic);
        if (result == null) {
            result = new MqttTopic(topic, mComms);
            mTopicMap.put(topic, result);
        }
        return result;
    }

    /*
     * 사용하지 않고 있음. 기본 TimerPingSender에서 주기적으로 발생시킴.
     * (non-Javadoc)
     * Check and send a ping if needed.
     * <p>By default, client sends PingReq to server to keep the connection to
     * server. For some platforms which cannot use this mechanism, such as Android,
     * developer needs to handle the ping request manually with this method.
     * </p>
     * @throws MqttException for other errors encountered while publishing the message.
     */
    public IMqttToken checkPing(Object userContext, IMqttActionListener callback) throws MqttException {
        MqttToken token;
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[ping] " + ">");
        }

        token = mComms.checkForActivity();
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[ping] " + "<");
        }

        return token;
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#subscribe(java.lang.String, int, java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken subscribe(String topicFilter, int qos, Object userContext, IMqttActionListener callback) throws MqttException {
        return this.subscribe(new String[]{topicFilter}, new int[]{qos}, userContext, callback);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#subscribe(java.lang.String, int)
     */
    @Override
    public IMqttToken subscribe(String topicFilter, int qos) throws MqttException {
        return this.subscribe(new String[]{topicFilter}, new int[]{qos}, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#subscribe(java.lang.String[], int[])
     */
    @Override
    public IMqttToken subscribe(String[] topicFilters, int[] qos) throws MqttException {
        return this.subscribe(topicFilters, qos, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#subscribe(java.lang.String[], int[], java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken subscribe(String[] topicFilters, int[] qos, Object userContext, IMqttActionListener callback) throws MqttException, IllegalArgumentException {
        if (topicFilters.length != qos.length) {
            throw new IllegalArgumentException();
        }

        StringBuilder builder = new StringBuilder();
        for (int i = 0; i < topicFilters.length; i++) {
            if (i > 0) {
                builder.append(", ");
            }
            builder.append(topicFilters[i] + ":" + qos[i]);

            //Check if the topic filter is valid before subscribing
            MqttTopic.validate(topicFilters[i], true/* allow wildcards */);
        }

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[subscribe] " + MessageFormat.format("Subscribe topic={0} userContext={1} callback={2}", new Object[]{builder.toString(), userContext, callback}));
        }

        MqttToken token = new MqttToken();
        token.setActionCallback(callback);
        token.setUserContext(userContext);
        token.internalTok.setTopics(topicFilters);

        MqttSubscribe register = new MqttSubscribe(topicFilters, qos);

        mComms.sendNoWait(register, token);
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[subscribe] " + "<");
        }

        return token;
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#unsubscribe(java.lang.String, java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken unsubscribe(String topicFilter, Object userContext, IMqttActionListener callback) throws MqttException {
        return unsubscribe(new String[]{topicFilter}, userContext, callback);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#unsubscribe(java.lang.String)
     */
    @Override
    public IMqttToken unsubscribe(String topicFilter) throws MqttException {
        return unsubscribe(new String[]{topicFilter}, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#unsubscribe(java.lang.String[])
     */
    @Override
    public IMqttToken unsubscribe(String[] topicFilters) throws MqttException {
        return unsubscribe(topicFilters, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#unsubscribe(java.lang.String[], java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttToken unsubscribe(String[] topicFilters, Object userContext, IMqttActionListener callback) throws MqttException, IllegalArgumentException {
        StringBuilder builder = new StringBuilder();
        for (int i = 0; i < topicFilters.length; i++) {
            if (i > 0) {
                builder.append(", ");
            }
            builder.append(topicFilters[i]);

            // Check if the topic filter is valid before unsubscribing
            // Although we already checked when subscribing, but invalid
            // topic filter is meanless for unsubscribing, just prohibit it
            // to reduce unnecessary control packet send to broker.
            MqttTopic.validate(topicFilters[i], true/* allow wildcards */);
        }

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[unsubscribe] " + MessageFormat.format("Unsubscribe topic={0} userContext={1} callback={2}", new Object[]{builder.toString(), userContext, callback}));
        }

        MqttToken token = new MqttToken();
        token.setActionCallback(callback);
        token.setUserContext(userContext);
        token.internalTok.setTopics(topicFilters);

        MqttUnsubscribe unregister = new MqttUnsubscribe(topicFilters);

        mComms.sendNoWait(unregister, token);
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[unsubscribe] " + "<");
        }

        return token;
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#setCallback(MqttCallback)
     */
    @Override
    public void setCallback(MqttCallback callback) {
        mComms.setCallback(callback);
    }

    /**
     * Returns a randomly generated client identifier based on the the fixed prefix (paho-)
     * and the system time.
     * <p>
     * When cleanSession is set to false, an application must ensure it uses the same client identifier when it reconnects to the server to resume state and maintain assured message delivery.
     * </p>
     *
     * @return a generated client identifier
     * @see MqttConnectOptions#setCleanSession(boolean)
     */
    public static String generateClientId() {
        //length of nanoTime = 15, so total length = 20  < 65535(defined in spec) 
        return CLIENT_ID_PREFIX + System.nanoTime();
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#getPendingDeliveryTokens()
     */
    @Override
    public IMqttDeliveryToken[] getPendingDeliveryTokens() {
        return mComms.getPendingDeliveryTokens();
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#publish(java.lang.String, byte[], int, boolean, java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttDeliveryToken publish(String topic, byte[] payload, int qos,
                                      boolean retained, Object userContext, IMqttActionListener callback) throws MqttException,
            MqttPersistenceException {
        MqttMessage message = new MqttMessage(payload);
        message.setQos(qos);
        message.setRetained(retained);
        return this.publish(topic, message, userContext, callback);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#publish(java.lang.String, byte[], int, boolean)
     */
    @Override
    public IMqttDeliveryToken publish(String topic, byte[] payload, int qos,
                                      boolean retained) throws MqttException, MqttPersistenceException {
        return this.publish(topic, payload, qos, retained, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#publish(java.lang.String, MqttMessage)
     */
    @Override
    public IMqttDeliveryToken publish(String topic, MqttMessage message) throws MqttException, MqttPersistenceException {
        return this.publish(topic, message, null, null);
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#publish(java.lang.String, MqttMessage, java.lang.Object, IMqttActionListener)
     */
    @Override
    public IMqttDeliveryToken publish(String topic, MqttMessage message, Object userContext, IMqttActionListener callback) throws MqttException,
            MqttPersistenceException, IllegalArgumentException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[publish] " + MessageFormat.format("< topic={0} message={1}userContext={1} callback={2}", new Object[]{topic, userContext, callback}));
        }

        //Checks if a topic is valid when publishing a message.
        MqttTopic.validate(topic, false/* wildcards NOT allowed */);

        MqttDeliveryToken token = new MqttDeliveryToken();
        token.setActionCallback(callback);
        token.setUserContext(userContext);
        token.setMessage(message);
        token.internalTok.setTopics(new String[]{topic});

        MqttPublish pubMsg = new MqttPublish(topic, message);
        mComms.sendNoWait(pubMsg, token);

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[publish] " + "<");
        }

        return token;
    }

    /*
     * (non-Javadoc)
     * @see IMqttAsyncClient#close()
     */
    @Override
    public void close() throws MqttException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[close] " + "<");
        }
        mComms.close();
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[close] " + ">");
        }
    }
}
