/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.comm.lib.utils.LogExUtils;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.IMqttAsyncClient;
import org.eclipse.paho.client.mqttv3.MqttCallback;
import org.eclipse.paho.client.mqttv3.MqttClientPersistence;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttPersistenceException;
import org.eclipse.paho.client.mqttv3.MqttPingSender;
import org.eclipse.paho.client.mqttv3.MqttToken;
import org.eclipse.paho.client.mqttv3.MqttTopic;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttConnack;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttConnect;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttDisconnect;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPublish;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttWireMessage;

import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Vector;

import androidx.annotation.NonNull;

/**
 * Handles client communications with the server. Sends and receives MQTT V3
 * messages.
 * 서버와의 통신을 다룬다.
 */
public class ClientComms {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "ClientComms";

    final static byte CONNECTED = 0;
    final static byte CONNECTING = 1;
    final static byte DISCONNECTING = 2;
    final static byte DISCONNECTED = 3;
    final static byte CLOSED = 4;

    private IMqttAsyncClient mAsyncClient;
    private NetworkModule mNetworkModule;
    /**
     * 수신 쓰레드.
     */
    CommsReceiver mCommsReceiver;
    /**
     * 송신 쓰레드.
     */
    private CommsSender mCommsSender;
    /**
     * 수신 처리 쓰레드.
     */
    private CommsCallback mCallback;
    /**
     * 핵심 중의 핵심.
     */
    private ClientState mClientState;
    /**
     * 연결 옵션.
     * MqttAsyncClient로부터 받음.
     */
    private MqttConnectOptions mConnectOptions;
    /**
     * 저장소.
     * MqttAsyncClient로부터 받음.
     */
    private MqttClientPersistence mPersistence;
    /**
     * MqttAsyncClient로부터 받음.
     */
    private MqttPingSender mPingSender;
    private CommsTokenStore mTokenStore;
    private boolean mIsDisconnecting = false;

    /**
     * 연결 상태.
     */
    private byte mConnectState = DISCONNECTED;

    /**
     * 락.
     * mConnectState를 보호하는 것은 확실하나 추가 용도가 불명확함.
     */
    private final Object mConnectLock = new Object();    // Used to synchronize connection state
    private boolean mClosePending = false;

    /**
     * (jaebok.lee)
     * DISCONNECT message를 보냈는지 여부.
     * send queue에 넣는 것까지 단계.
     */
    private boolean mDisconnectSent = false;

    /**
     * Creates a new ClientComms object, using the specified module to handle
     * the network calls.
     */
    public ClientComms(IMqttAsyncClient asyncClient, MqttClientPersistence persistence, MqttPingSender pingSender) throws MqttException {
        this.mConnectState = DISCONNECTED;
        this.mAsyncClient = asyncClient;
        this.mPersistence = persistence;
        this.mPingSender = pingSender;
        this.mPingSender.init(this);

        this.mTokenStore = new CommsTokenStore();
        this.mCallback = new CommsCallback(this);
        this.mClientState = new ClientState(persistence, mTokenStore, this.mCallback, this, pingSender);

        mCallback.setClientState(mClientState);
    }

    /**
     * Sends a message to the server. Does not check if connected this validation must be done
     * by invoking routines.
     * 여러 쓰레드에서 호출.
     *
     * @param message
     * @param token
     * @throws MqttException
     */
    /*package*/ void internalSend(@NonNull MqttWireMessage message, @NonNull MqttToken token) throws MqttException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[internalSend] " + MessageFormat.format("internalSend key={0} message={1} token={2}", new Object[]{message.getKey(), message, token}));
        }

        if (token.getClient() == null) {
            // Associate the mAsyncClient with the token - also marks it as in use.
            token.internalTok.setClient(mAsyncClient);
        } else {
            // client가 있다는 것은 사용중이라는 의미.
            // Token is already in use - cannot reuse
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[internalSend]" + " " + MessageFormat.format("fail: token in use: key={0} message={1} token={2}", new Object[]{message.getKey(), message, token}));
            }

            throw new MqttException(MqttException.REASON_CODE_TOKEN_INUSE);
        }

        try {
            // Persist if needed and send the message
            mClientState.send(message, token);
        } catch (MqttException e) {
            if (message instanceof MqttPublish) {
                mClientState.undo((MqttPublish) message);
            }
            throw e;
        }
    }

    /**
     * Sends a message to the broker if in connected state, but only waits for the message to be
     * stored, before returning.
     */
    public void sendNoWait(MqttWireMessage message, MqttToken token) throws MqttException {
        if (isConnected() ||
                message instanceof MqttConnect ||
                isDisconnecting() && message instanceof MqttDisconnect) {
            internalSend(message, token);
        } else {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[sendNoWait] " + "failed: not connected");
            }
            throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_NOT_CONNECTED);
        }
    }

    /**
     * Close and tidy up.
     * <p/>
     * Call each main class and let it tidy up e.g. releasing the token
     * store which normally survives a disconnect.
     *
     * @throws MqttException if not disconnected
     */
    public void close() throws MqttException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[close]");
        }
        synchronized (mConnectLock) {
            if (!isClosed()) {
                // Must be disconnected before close can take place
                if (!isDisconnected()) {
                    if (DEBUG) {
                        LogEx.v(LOG_TAG, "[close] " + "failed: not disconnected");
                    }

                    if (isConnecting()) {
                        throw new MqttException(MqttException.REASON_CODE_CONNECT_IN_PROGRESS);
                    } else if (isConnected()) {
                        throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_CONNECTED);
                    } else if (isDisconnecting()) {
                        mClosePending = true;
                        return;
                    }
                }

                mConnectState = CLOSED;

                // ShutdownConnection has already cleaned most things
                mClientState.close();
                mClientState = null;
                mCallback = null;
                mPersistence = null;
                mCommsSender = null;
                mPingSender = null;
                mCommsReceiver = null;
                mNetworkModule = null;
                mConnectOptions = null;
                mTokenStore = null;
            }
        }
    }

    /**
     * Sends a connect message and waits for an ACK or NACK.
     * Connecting is a special case which will also start up the
     * network connection, receive thread, and keep alive thread.
     */
    public void connect(MqttConnectOptions options, MqttToken token) throws MqttException {
        synchronized (mConnectLock) {
            if (isDisconnected() && !mClosePending) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[connect] " + "state=CONNECTING");
                }

                mConnectState = CONNECTING;

                this.mConnectOptions = options;

                // connect 메시지 생성.
                MqttConnect connect = new MqttConnect(mAsyncClient.getClientId(),
                        options.isCleanSession(),
                        options.getKeepAliveInterval(),
                        options.getUserName(),
                        options.getPassword(),
                        options.getWillMessage(),
                        options.getWillDestination());

                this.mClientState.setKeepAliveSecs(options.getKeepAliveInterval());
                this.mClientState.setCleanSession(options.isCleanSession());

                mTokenStore.open();
                ConnectBG connectThread = new ConnectBG(this, token, connect);
                connectThread.start();
            } else {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[connect] " + MessageFormat.format("connect failed: not disconnected {0}", new Object[]{Byte.valueOf(mConnectState)}));
                }
                if (isClosed() || mClosePending) {
                    throw new MqttException(MqttException.REASON_CODE_CLIENT_CLOSED);
                } else if (isConnecting()) {
                    throw new MqttException(MqttException.REASON_CODE_CONNECT_IN_PROGRESS);
                } else if (isDisconnecting()) {
                    throw new MqttException(MqttException.REASON_CODE_CLIENT_DISCONNECTING);
                } else {
                    throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_CONNECTED);
                }
            }
        }
    }

    /**
     * CONNECT 요청에 대한 응답 시 호출.
     *
     * @param connack
     */
    public void connectComplete(MqttConnack connack) {
        int rc = connack.getReturnCode();
        synchronized (mConnectLock) {
            if (rc == 0) {
                // We've successfully connected
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[connectComplete] " + "state=CONNECTED");
                }

                mConnectState = CONNECTED;
                return;
            }
        }

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[connectComplete] " + MessageFormat.format("connect failed: rc={0}", new Object[]{Integer.valueOf(rc)}));

        }
    }

    /**
     * Shuts down the connection to the server.
     * This may have been invoked as a result of a user calling disconnect or
     * an abnormal disconnection. The method may be invoked multiple times
     * in parallel as each thread when it receives an error uses this method
     * to ensure that shutdown completes successfully.
     */
    public void shutdownConnection(MqttToken token, MqttException reason) {
        boolean wasConnected;
        MqttToken endToken;        //Token to notify after disconnect completes

        // This method could concurrently be invoked from many places only allow it
        // to run once.
        synchronized (mConnectLock) {
            if (mIsDisconnecting || mClosePending) {
                return;
            }
            mIsDisconnecting = true;

            if (DEBUG) {
                LogEx.v(LOG_TAG, "[shutdownConnection] " + "state=DISCONNECTING");

            }

            wasConnected = isConnected() || isDisconnecting();
            mConnectState = DISCONNECTING;
        }

        // Update the token with the reason for shutdown if it
        // is not already complete.
        if (token != null && !token.isComplete()) {
            token.internalTok.setException(reason);
        }

        // Stop the thread that is used to call the user back
        // when actions complete
        if (mCallback != null) {
            mCallback.stop();
        }

        // Stop the network module, send and receive now not possible
        try {
            if (mNetworkModule != null) {
                mNetworkModule.stop();
            }
        } catch (Exception ioe) {
            // Ignore as we are shutting down
        }

        // Stop the thread that handles inbound work from the network
        if (mCommsReceiver != null) {
            mCommsReceiver.stop();
        }

        // Stop any new tokens being saved by app and throwing an exception if they do
        mTokenStore.quiesce(new MqttException(MqttException.REASON_CODE_CLIENT_DISCONNECTING));

        // Notify any outstanding tokens with the exception of
        // con or discon which may be returned and will be notified at
        // the end
        // 연결 / 해제 토큰 리턴 받음.
        endToken = handleOldTokens(token, reason);
        if (DEBUG) {
            LogEx.d(LOG_TAG, "[endToken] endToken : " + endToken);
        }

        try {
            // Clean session handling and tidy up
            mClientState.disconnected(reason);
        } catch (Exception ex) {
            // Ignore as we are shutting down
        }

        if (mCommsSender != null) {
            mCommsSender.stop();
        }

        if (mPingSender != null) {
            mPingSender.stop();
        }

        try {
            if (mPersistence != null) {
                mPersistence.close();
            }
        } catch (Exception ex) {
            // Ignore as we are shutting down
        }
        // All disconnect logic has been completed allowing the
        // mAsyncClient to be marked as disconnected.
        synchronized (mConnectLock) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[shutdownConnection] " + "state=DISCONNECTED");
            }

            mConnectState = DISCONNECTED;
            mIsDisconnecting = false;
        }

        // Internal disconnect processing has completed.  If there
        // is a disconnect token or a connect in error notify
        // it now. This is done at the end to allow a new connect
        // to be processed and now throw a currently disconnecting error.
        // any outstanding tokens and unblock any waiters
        if (endToken != null & mCallback != null) {
            mCallback.asyncOperationComplete(endToken);
        }

        if (wasConnected && mCallback != null) {
            // Let the user know mAsyncClient has disconnected either normally or abnormally
            mCallback.connectionLost(reason);
        }

        // While disconnecting, close may have been requested - try it now
        synchronized (mConnectLock) {
            if (mClosePending) {
                try {
                    close();
                } catch (Exception e) { // ignore any errors as closing
                }
            }
        }
    }

    // shutdown 시 호출.
    // 토큰. 연결/해제 토큰을 두고 나머지 Complete queue로 전달.
    // 연결/해제 토큰은 리턴.
    // Tidy up. There may be tokens outstanding as the mAsyncClient was
    // not disconnected/quiseced cleanly! Work out what tokens still
    // need to be notified and waiters unblocked. Store the
    // disconnect or connect token to notify after disconnect is
    // complete.
    private MqttToken handleOldTokens(MqttToken token, MqttException reason) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[handleOldTokens] " + ">");
        }

        MqttToken tokenToNotifyLater = null;
        try {
            // First the token that was related to the disconnect / shutdown may
            // not be in the token table - temporarily add it if not
            if (token != null) {
                if (DEBUG) {
                    LogEx.d(LOG_TAG, "[handleOldTokens] token.internalTok.getKey() : " + token.internalTok.getKey());
                }
                if (mTokenStore.getToken(token.internalTok.getKey()) == null) {
                    mTokenStore.saveToken(token, token.internalTok.getKey());
                }
            }

            Vector<MqttToken> tokensToNotify = mClientState.resolveOldTokens(reason);
            Enumeration<MqttToken> tokensToNotifyEnumeration = tokensToNotify.elements();
            while (tokensToNotifyEnumeration.hasMoreElements()) {
                MqttToken tok = tokensToNotifyEnumeration.nextElement();

                if (tok.internalTok.getKey().equals(MqttDisconnect.KEY) ||
                        tok.internalTok.getKey().equals(MqttConnect.KEY)) {
                    // Its connect or disconnect so remember and notify @ end of disc routine
                    tokenToNotifyLater = tok;
                } else {
                    // notify waiters and callbacks of outstanding tokens
                    // that a problem has occurred and disconnect is in
                    // progress
                    mCallback.asyncOperationComplete(tok);
                }
            }
        } catch (Exception ex) {
            // Ignore as we are shutting down

            if (LogExUtils.isLogExWarn()) {
                LogEx.w(LOG_TAG, "[handleOldTokens] exception", ex);
            }
        }
        return tokenToNotifyLater;
    }

    /**
     * 연결 끊기 쓰레드 실행하고 나옴.
     *
     * @param disconnect
     * @param quiesceTimeout
     * @param token          사용자 토큰.
     * @throws MqttException
     */
    public void disconnect(MqttDisconnect disconnect, long quiesceTimeout, MqttToken token) throws MqttException {
        synchronized (mConnectLock) {
            if (isClosed()) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[disconnect] " + "failed: in closed state");
                }
                throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_CLOSED);
            } else if (isDisconnected()) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[disconnect] " + "failed: already disconnected");
                }
                throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_ALREADY_DISCONNECTED);
            } else if (isDisconnecting()) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[disconnect] " + "failed: already disconnecting");
                }
                throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_DISCONNECTING);
            } else if (Thread.currentThread() == mCallback.getThread()) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[disconnect] " + "failed: called on callback thread");
                }
                // Not allowed to call disconnect() from the callback, as it will deadlock.
                throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_DISCONNECT_PROHIBITED);
            }

            if (DEBUG) {
                LogEx.v(LOG_TAG, "[disconnect] " + "state=DISCONNECTING");
            }
            mConnectState = DISCONNECTING;
            DisconnectBG discbg = new DisconnectBG(disconnect, quiesceTimeout, token);
            discbg.start();
        }
    }

    /**
     * Disconnect the connection and reset all the states.
     */
    public void disconnectForcibly(long quiesceTimeout, long disconnectTimeout) throws MqttException {
        // Allow current inbound and outbound work to complete
        mClientState.quiesce(quiesceTimeout);
        MqttToken token = new MqttToken();
        try {
            // Send disconnect packet
            internalSend(new MqttDisconnect(), token);
            mDisconnectSent = true;

            // Wait util the disconnect packet sent with timeout
            token.waitForCompletion(disconnectTimeout);
        } catch (Exception ex) {
            // ignore, probably means we failed to send the disconnect packet.
        } finally {
            token.internalTok.markComplete(null, null);
            shutdownConnection(token, null);
        }
    }

    public boolean isConnected() {
        return mConnectState == CONNECTED;
    }

    public boolean isConnecting() {
        return mConnectState == CONNECTING;
    }

    public boolean isDisconnected() {
        return mConnectState == DISCONNECTED;
    }

    public boolean isDisconnecting() {
        return mConnectState == DISCONNECTING;
    }

    public boolean isClosed() {
        return mConnectState == CLOSED;
    }

    public boolean isDisconnectSent() {
        return mDisconnectSent;
    }

    public void setCallback(MqttCallback mqttCallback) {
        this.mCallback.setCallback(mqttCallback);
    }

    protected MqttTopic getTopic(String topic) {
        return new MqttTopic(topic, this);
    }

    public NetworkModule getNetworkModule() {
        return mNetworkModule;
    }

    public void setNetworkModule(NetworkModule networkModule) {
        this.mNetworkModule = networkModule;
    }

    public MqttDeliveryToken[] getPendingDeliveryTokens() {
        return mTokenStore.getOutstandingDelTokens();
    }

    protected void deliveryComplete(MqttPublish msg) throws MqttPersistenceException {
        mClientState.deliveryComplete(msg);
    }

    public IMqttAsyncClient getAsyncClient() {
        return mAsyncClient;
    }

    public long getKeepAlive() {
        return this.mClientState.getKeepAlive();
    }

    public ClientState getClientState() {
        return mClientState;
    }

    public MqttConnectOptions getConOptions() {
        return mConnectOptions;
    }

    // Kick off the connect processing in the background so that it does not block. For instance
    // the socket could take time to create.
    private class ConnectBG implements Runnable {
        ClientComms mClientComms = null;
        Thread mThread = null;
        MqttToken mToken;
        MqttConnect mMessage;

        ConnectBG(ClientComms clientComms, MqttToken token, MqttConnect connect) {
            mClientComms = clientComms;
            mToken = token;
            mMessage = connect;
            mThread = new Thread(this, "MQTT Con: " + getAsyncClient().getClientId());
        }

        void start() {
            mThread.start();
        }

        @Override
        public void run() {
            MqttException mqttEx = null;
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[connectBG:run] " + ">");
            }

            try {
                // Reset an exception on existing delivery tokens.
                // This will have been set if disconnect occured before delivery was
                // fully processed.
                MqttDeliveryToken[] toks = mTokenStore.getOutstandingDelTokens();
                for (int i = 0; i < toks.length; i++) {
                    toks[i].internalTok.setException(null);
                }

                // Save the connect token in tokenStore as failure can occur before send
                mTokenStore.saveToken(mToken, mMessage);

                // Connect to the server at the network level e.g. TCP socket and then
                // start the background processing threads before sending the connect
                // packet.
                // 소켓 생성 및 연결.
                mNetworkModule.start();
                // input 연결 쓰레드 시작.
                mCommsReceiver = new CommsReceiver(mClientComms, mClientState, mTokenStore, mNetworkModule.getInputStream());
                mCommsReceiver.start("MQTT Rec: " + getAsyncClient().getClientId());
                // output 연결 쓰레드 시작.
                mCommsSender = new CommsSender(mClientComms, mClientState, mTokenStore, mNetworkModule.getOutputStream());
                mCommsSender.start("MQTT Snd: " + getAsyncClient().getClientId());
                // 콜백 쓰레드 시작.
                mCallback.start("MQTT Call: " + getAsyncClient().getClientId());
                internalSend(mMessage, mToken);
            } catch (MqttException ex) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[connectBG:run] " + "connect failed: unexpected exception");
                }
                mqttEx = ex;
            } catch (Exception ex) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[connectBG:run] " + "connect failed: unexpected exception");
                }
                mqttEx = ExceptionHelper.createMqttException(ex);
            }

            if (mqttEx != null) {
                shutdownConnection(mToken, mqttEx);
            }
        }
    }

    // Kick off the disconnect processing in the background so that it does not block. For instance
    // the quiesce
    private class DisconnectBG implements Runnable {
        private Thread mThread = null;
        private final MqttDisconnect mMessage;
        private final long mQuiesceTimeout;
        private final MqttToken mToken;

        DisconnectBG(MqttDisconnect disconnect, long quiesceTimeout, MqttToken token) {
            mMessage = disconnect;
            mQuiesceTimeout = quiesceTimeout;
            mToken = token;
        }

        void start() {
            mThread = new Thread(this, "MQTT Disc: " + getAsyncClient().getClientId());
            mThread.start();
        }

        @Override
        public void run() {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[disconnectBG:run] " + ">");
            }

            // Allow current inbound and outbound work to complete
            mClientState.quiesce(mQuiesceTimeout);
            try {
                internalSend(mMessage, mToken);
                mToken.internalTok.waitUntilSent();
            } catch (MqttException ex) {
                // (jaebok.lee) ignore ?
            } finally {
                mToken.internalTok.markComplete(null, null);
                shutdownConnection(mToken, null);
            }
        }
    }

    /*
     * 타이머 쓰레드에서 호출.
     * Check and send a ping if needed and check for ping timeout.
     * Need to send a ping if nothing has been sent or received
     * in the last keepalive interval.
     */
    public MqttToken checkForActivity() {
        MqttToken token = null;
        try {
            token = mClientState.checkForActivity();
        } catch (Exception e) {
            handleRunException(e);
        }
        return token;
    }

    /**
     * 연결 끊기.
     *
     * @param ex
     */
    private void handleRunException(Exception ex) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[handleRunException] " + "exception");
        }
        MqttException mex;
        if ((ex instanceof MqttException) == false) {
            mex = new MqttException(MqttException.REASON_CODE_CONNECTION_LOST, ex);
        } else {
            mex = (MqttException) ex;
        }

        shutdownConnection(null, mex);
    }
}
