/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.comm.lib.utils.LogExUtils;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.MqttClientPersistence;
import org.eclipse.paho.client.mqttv3.MqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.eclipse.paho.client.mqttv3.MqttPersistable;
import org.eclipse.paho.client.mqttv3.MqttPersistenceException;
import org.eclipse.paho.client.mqttv3.MqttPingSender;
import org.eclipse.paho.client.mqttv3.MqttToken;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttAck;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttConnack;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttConnect;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPingReq;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPingResp;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPubAck;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPubComp;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPubRec;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPubRel;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPublish;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttWireMessage;

import java.io.EOFException;
import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import androidx.annotation.NonNull;

/**
 * 미 전달 메시지들의 상태 정보 보관.
 * The core of the client, which holds the state information for pending and in-flight messages.
 * <p/>
 * Messages that have been accepted for delivery are moved between several objects while being delivered.
 * <p/>
 * 1) When the client is not running messages are stored in a persistent store that implements the MqttClientPersistent Interface. The default is MqttDefaultFilePersistencew which stores messages
 * safely across failures and system restarts. If no persistence is specified there is a fall back to MemoryPersistence which will maintain the messages while the Mqtt client is instantiated.
 * <p/>
 * 2) When the client or specifically ClientState is instantiated the messages are read from the persistent store into: - outboundqos2 hashtable if a QoS 2 PUBLISH or PUBREL - outboundqos1 hashtable
 * if a QoS 1 PUBLISH (see restoreState)
 * <p/>
 * 3) On Connect, copy messages from the outbound hashtables to the mPendingMessages or mPendingFlowQueue vector in message id order. - Initial message publish goes onto the pending messages buffer. - PUBREL
 * goes onto the pending flows buffer {@see restoreInflightMessages}
 * <p/>
 * 4) Sender thread reads messages from the pendingflows and pendingmessages buffer one at a time. The message is removed from the pendingbuffer but remains on the outbound* hashtable. The hashtable
 * is the place where the full set of outstanding messages are stored in memory. (Persistence is only used at start up)
 * <p/>
 * 5) Receiver thread - receives wire messages: - if QoS 1 then remove from persistence and outboundqos1 - if QoS 2 PUBREC send PUBREL. Updating the outboundqos2 entry with the PUBREL and update
 * persistence. - if QoS 2 PUBCOMP remove from persistence and outboundqos2
 * <p/>
 * Notes: because of the multithreaded nature of the client it is vital that any changes to this class take concurrency into account. For instance as soon as a flow / message is put on the wire it is
 * possible for the receiving thread to receive the ack and to be processing the response before the sending side has finished processing. For instance a connect may be sent, the conack received
 * before the connect notify send has been processed!
 */
public class ClientState {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "ClientState";

    public static final int PING_TIMEOUT_IN_MILLS = 5 * 1000;


    private static final String PERSISTENCE_SENT_PREFIX = "s-";
    private static final String PERSISTENCE_CONFIRMED_PREFIX = "sc-";
    private static final String PERSISTENCE_RECEIVED_PREFIX = "r-";

    private static final int MAX_INFLIGHT = 10;


    private static final int MIN_MSG_ID = 1;        // Lowest possible MQTT message ID to use
    private static final int MAX_MSG_ID = 65535;    // Highest possible MQTT message ID to use
    private int mNextMsgId = MIN_MSG_ID - 1;            // The next available message ID to use


    /**
     * 사용중인 message id 관리.
     * id 발급 시 사용.
     */
    private Hashtable<Integer, Integer> mInUseMsgIdMap;                    // Used to store a set of in-use message IDs

    /**
     * 송신 메시지 큐.
     * publish 용도.
     * 메시지 얻을 경우 mPendingFlows보다 늦게 얻음.
     */
    volatile private Vector<MqttWireMessage> mPendingMessages;
    /**
     * 송신 메시지 큐.
     * publish 아닌 용도.
     * connect 경우 맨 앞 위치.
     * 메시지 얻을 경우 mPendingMessages보다 먼저 얻음.
     */
    volatile private Vector<MqttWireMessage> mPendingFlowQueue;

    /**
     * 송신 메시지에 대한 저장 관리.
     * 추적 용도.
     */
    private CommsTokenStore mTokenStore;

    private ClientComms mClientComms = null;


    private CommsCallback mCallbackThread = null;

    /**
     * keepalive 시간.
     */
    private long mKeepAliveInMillis;

    private boolean mCleanSession;

    /**
     * Qos1, 2를 위한 저장 관리.
     */
    private MqttClientPersistence mPersistence;

    /**
     * 처리중인 PUBLISH 메시지 개수.
     * MAX_INFLIGHT로 제한.
     */
    private int mActualInFlightCount = 0;

    /**
     * 처리중인(MQTT PUBCOMP 받기 전까지의) MQTT PUBREL 메시지 개수.
     * 디버그용.
     */
    private int mInFlightPubRelCount = 0;

    /**
     * 일단 송신 메시지 큐를 보호한다는 것은 분명하나 그외 용도 불명확.
     */
    private final Object mQueueLock = new Object();
    private final Object mQuiesceLock = new Object();
    private boolean mQuiescing = false;

    /**
     * 총 메시지 통틀어 마지막 보낸 시각.
     * [핑 관련]
     */
    private long mLastOutboundActivityTimeInMillis = 0;
    /**
     * 총 메시지 통틀어 마지막 받은 시각.
     * [핑 관련]
     */
    private long mLastInboundActivityInMillis = 0;
    /**
     * 마지막 PINGREQ 보낸 시각.
     * [핑 관련]
     */
    private long mLastPingTimeInMillis = 0;
    /**
     * PINGREQ 메시지. 미리 만들어 재사용.
     * [핑 관련]
     */
    private MqttWireMessage mPingReqMessage;
    /**
     * PINGREQ 메시지가 보내지고 돌아오지 않은 상태인지 여부.
     * [핑 관련]
     * 자체 락 보호.
     */
    private Boolean mIsPingOutstanding = Boolean.FALSE;

    private final Object mPingLock = new Object();

    /**
     * 핑 처리를 위한 연결 상태.
     */
    private boolean mConnected = false;

    private Hashtable<Integer, MqttWireMessage> mOutboundQoS2Map = null;
    private Hashtable<Integer, Object> mOutboundQoS1Map = null;
    private Hashtable<Integer, Object> mInboundQoS2Map = null;

    private MqttPingSender mPingSender = null;

    protected ClientState(MqttClientPersistence persistence, CommsTokenStore tokenStore,
                          CommsCallback callback, ClientComms clientComms, MqttPingSender pingSender) throws MqttException {

        mInUseMsgIdMap = new Hashtable<Integer, Integer>();
        mPendingMessages = new Vector<MqttWireMessage>(MAX_INFLIGHT);
        mPendingFlowQueue = new Vector<MqttWireMessage>();
        mOutboundQoS2Map = new Hashtable<Integer, MqttWireMessage>();
        mOutboundQoS1Map = new Hashtable<Integer, Object>();
        mInboundQoS2Map = new Hashtable<Integer, Object>();
        mPingReqMessage = new MqttPingReq();
        mInFlightPubRelCount = 0;
        mActualInFlightCount = 0;

        this.mPersistence = persistence;
        this.mCallbackThread = callback;
        this.mTokenStore = tokenStore;
        this.mClientComms = clientComms;
        this.mPingSender = pingSender;

        restoreState();
    }

    protected void setKeepAliveSecs(long keepAliveSecs) {
        this.mKeepAliveInMillis = keepAliveSecs * 1000;
    }

    protected long getKeepAlive() {
        return this.mKeepAliveInMillis;
    }

    protected void setCleanSession(boolean cleanSession) {
        this.mCleanSession = cleanSession;
    }

    private static String getSendPersistenceKey(@NonNull MqttWireMessage message) {
        return PERSISTENCE_SENT_PREFIX + message.getMessageId();
    }

    private static String getSendConfirmPersistenceKey(@NonNull MqttWireMessage message) {
        return PERSISTENCE_CONFIRMED_PREFIX + message.getMessageId();
    }

    private static String getReceivedPersistenceKey(@NonNull MqttWireMessage message) {
        return PERSISTENCE_RECEIVED_PREFIX + message.getMessageId();
    }

    /**
     * TODO lock 보호 필요.
     *
     * @throws MqttException
     */
    private void clearState() throws MqttException {
        if (DEBUG) {
            LogEx.d(LOG_TAG, "[clearState] " + "clearState");
        }

        mPersistence.clear();
        mInUseMsgIdMap.clear();
        mPendingMessages.clear();
        mPendingFlowQueue.clear();
        mOutboundQoS2Map.clear();
        mOutboundQoS1Map.clear();
        mInboundQoS2Map.clear();
        mTokenStore.clear();
    }

    /**
     * restoreState에서 호출.
     *
     * @param key
     * @param persistable
     * @return
     * @throws MqttException
     */
    private MqttWireMessage restoreMessage(String key, MqttPersistable persistable) throws MqttException {
        MqttWireMessage message = null;

        try {
            message = MqttWireMessage.createWireMessage(persistable);
        } catch (MqttException ex) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[restoreMessage] " + MessageFormat.format("key={0} exception", new Object[]{key}));
            }
            if (ex.getCause() instanceof EOFException) {
                // Premature end-of-file means that the message is corrupted
                if (key != null) {
                    mPersistence.remove(key);
                }
            } else {
                throw ex;
            }
        }
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[restoreMessage] " + MessageFormat.format("key={0} message={1}", new Object[]{key, message}));
        }
        return message;
    }

    /**
     * message id를 가지고 정렬 고려하여 삽입. (오름차순)
     * mQueueLock 내에서 호출.
     * Inserts a new message to the list, ensuring that list is ordered from lowest to highest in terms of the message id's.
     *
     * @param list   the list to insert the message into
     * @param newMsg the message to insert into the list
     */
    private static void insertInOrder(Vector<MqttWireMessage> list, MqttWireMessage newMsg) {
        int newMsgId = newMsg.getMessageId();
        for (int i = 0; i < list.size(); i++) {
            MqttWireMessage otherMsg = list.elementAt(i);
            int otherMsgId = otherMsg.getMessageId();
            if (otherMsgId > newMsgId) {
                list.insertElementAt(newMsg, i);
                return;
            }
        }
        list.addElement(newMsg);
    }

    /**
     * Produces a new list with the messages properly ordered according to their message id's.
     * mQueueLock 내에서 호출.
     *
     * @param list the list containing the messages to produce a new reordered list for
     *             - this will not be modified or replaced, i.e., be read-only to this method
     * @return a new reordered list
     */
    private static Vector<MqttWireMessage> reOrder(Vector<MqttWireMessage> list) {

        // here up the new list
        Vector<MqttWireMessage> newList = new Vector<MqttWireMessage>();

        if (list.size() == 0) {
            return newList; // nothing to reorder
        }

        // 루프 내 이전 message id.
        int previousMsgId = 0;
        int largestGap = 0;
        int largestGapMsgIdPosInList = 0;
        for (int i = 0; i < list.size(); i++) {
            // 루프 돌며 gap 가장 큰 위치 얻기.
            int currentMsgId = list.elementAt(i).getMessageId();
            if (currentMsgId - previousMsgId > largestGap) {
                largestGap = currentMsgId - previousMsgId;
                largestGapMsgIdPosInList = i;
            }
            previousMsgId = currentMsgId;
        }
        int firstMsgId = list.elementAt(0).getMessageId();
        int lastMsgId = previousMsgId; // last in the sorted list

        // we need to check that the gap after highest msg id to the lowest msg id is not beaten
        if (MAX_MSG_ID - lastMsgId + firstMsgId > largestGap) {
            largestGapMsgIdPosInList = 0;
        }

        // starting message has been located, let's start from this point on
        for (int i = largestGapMsgIdPosInList; i < list.size(); i++) {
            newList.addElement(list.elementAt(i));
        }

        // and any wrapping back to the beginning
        for (int i = 0; i < largestGapMsgIdPosInList; i++) {
            newList.addElement(list.elementAt(i));
        }

        return newList;
    }

    /**
     * 생성자에서 호출.
     * Restores the state information from persistence.
     */
    protected void restoreState() throws MqttException {
        Enumeration<String> messageKeys = mPersistence.keys();
        MqttPersistable persistable;
        String key;
        int highestMsgId = mNextMsgId;
        Vector<String> orphanedPubRels = new Vector<String>();
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[restoreMessage] " + ">");
        }

        while (messageKeys.hasMoreElements()) {
            key = messageKeys.nextElement();
            persistable = mPersistence.get(key);
            MqttWireMessage message = restoreMessage(key, persistable);
            if (message != null) {
                if (key.startsWith(PERSISTENCE_RECEIVED_PREFIX)) {
                    if (DEBUG) {
                        LogEx.v(LOG_TAG, "[restoreMessage] " + MessageFormat.format("inbound QoS 2 publish key={0} message={1}", new Object[]{key, message}));
                    }

                    // The inbound messages that we have persisted will be QoS 2
                    mInboundQoS2Map.put(Integer.valueOf(message.getMessageId()), message);
                } else if (key.startsWith(PERSISTENCE_SENT_PREFIX)) {
                    MqttPublish sendMessage = (MqttPublish) message;
                    highestMsgId = Math.max(sendMessage.getMessageId(), highestMsgId);
                    if (mPersistence.containsKey(getSendConfirmPersistenceKey(sendMessage))) {
                        MqttPersistable persistedConfirm = mPersistence.get(getSendConfirmPersistenceKey(sendMessage));
                        // QoS 2, and CONFIRM has already been sent...
                        MqttPubRel confirmMessage = (MqttPubRel) restoreMessage(key, persistedConfirm);
                        if (confirmMessage != null) {
                            confirmMessage.setDuplicate(true);
                            if (DEBUG) {
                                LogEx.v(LOG_TAG, "[restoreState] " + MessageFormat.format("outbound QoS 2 pubrel key={0} message={1}", new Object[]{key, message}));
                            }

                            mOutboundQoS2Map.put(Integer.valueOf(confirmMessage.getMessageId()), confirmMessage);
                        } else {
                            if (DEBUG) {
                                LogEx.v(LOG_TAG, "[restoreState] " + MessageFormat.format("outbound QoS 2 completed key={0} message={1}", new Object[]{key, message}));
                            }
                        }
                    } else {
                        // QoS 1 or 2, with no CONFIRM sent...
                        // Put the SEND to the list of pending messages, ensuring message ID ordering...
                        sendMessage.setDuplicate(true);
                        if (sendMessage.getMessage().getQos() == 2) {
                            if (DEBUG) {
                                LogEx.v(LOG_TAG, "[restoreState] " + MessageFormat.format("outbound QoS 2 publish key={0} message={1}", new Object[]{key, message}));
                            }

                            mOutboundQoS2Map.put(Integer.valueOf(sendMessage.getMessageId()), sendMessage);
                        } else {
                            if (DEBUG) {
                                LogEx.v(LOG_TAG, "[restoreState] " + MessageFormat.format("outbound QoS 1 publish key={0} message={1}", new Object[]{key, message}));
                            }

                            mOutboundQoS1Map.put(Integer.valueOf(sendMessage.getMessageId()), sendMessage);
                        }
                    }
                    MqttDeliveryToken tok = mTokenStore.restoreToken(sendMessage);
                    tok.internalTok.setClient(mClientComms.getAsyncClient());
                    mInUseMsgIdMap.put(Integer.valueOf(sendMessage.getMessageId()), Integer.valueOf(sendMessage.getMessageId()));
                } else if (key.startsWith(PERSISTENCE_CONFIRMED_PREFIX)) {
                    MqttPubRel pubRelMessage = (MqttPubRel) message;
                    if (!mPersistence.containsKey(getSendPersistenceKey(pubRelMessage))) {
                        orphanedPubRels.addElement(key);
                    }
                }
            }
        }

        messageKeys = orphanedPubRels.elements();
        while (messageKeys.hasMoreElements()) {
            key = messageKeys.nextElement();
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[restoreState] " + MessageFormat.format("removing orphaned pubrel key={0}", new Object[]{key}));
            }

            mPersistence.remove(key);
        }

        mNextMsgId = highestMsgId;
    }

    /**
     * mOutboundQoS2Map, mOutboundQoS1Map로부터 mPendingMessages, mPendingFlowQueue 재구성.
     * CONNECT 응답 온 경우에 호출.
     * CleanSession true 경우 호출은 되지만 무의미.
     * connect 성공 시 mQueueLock 내에서 호출.
     */
    private void restoreInflightMessages() {
        mPendingMessages = new Vector<MqttWireMessage>(MAX_INFLIGHT);
        mPendingFlowQueue = new Vector<MqttWireMessage>();

        // Qos2
        Enumeration<Integer> keys = mOutboundQoS2Map.keys();
        while (keys.hasMoreElements()) {
            Integer key = keys.nextElement();
            Object msg = mOutboundQoS2Map.get(key);
            if (msg instanceof MqttPublish) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[restoreInflightMessages] " + MessageFormat.format("QoS 2 publish key={0}", new Object[]{key}));
                }

                insertInOrder(mPendingMessages, (MqttPublish) msg);
            } else if (msg instanceof MqttPubRel) {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[restoreInflightMessages] " + MessageFormat.format("QoS 2 pubrel key={0}", new Object[]{key}));
                }

                insertInOrder(mPendingFlowQueue, (MqttPubRel) msg);
            }
        }
        // Qos1
        keys = mOutboundQoS1Map.keys();
        while (keys.hasMoreElements()) {
            Integer key = keys.nextElement();
            MqttPublish msg = (MqttPublish) mOutboundQoS1Map.get(key);
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[restoreInflightMessages] " + MessageFormat.format("QoS 1 publish key={0}", new Object[]{key}));
            }

            insertInOrder(mPendingMessages, msg);
        }

        this.mPendingFlowQueue = reOrder(mPendingFlowQueue);
        this.mPendingMessages = reOrder(mPendingMessages);
    }

    /**
     * Submits a message for delivery. This method will block until there is
     * room in the inFlightWindow for the message. The message is put into
     * persistence before returning.
     * 여러 쓰레드에서 호출 진입.
     *
     * @param message the message to send
     * @param token   the token that can be used to track delivery of the message
     * @throws MqttException
     */
    public void send(MqttWireMessage message, MqttToken token) throws MqttException {
        if (message.isMessageIdRequired() && message.getMessageId() == 0) {
            message.setMessageId(getNextMessageId());
        }
        if (token != null) {
            try {
                token.internalTok.setMessageID(message.getMessageId());
            } catch (Exception ignored) {
            }
        }

        //
        // publish와 그외 경우로 나뉜다.
        //
        if (message instanceof MqttPublish) {
            synchronized (mQueueLock) {
                if (mActualInFlightCount >= MAX_INFLIGHT) {
                    if (DEBUG) {
                        LogEx.v(LOG_TAG, "[send] " + MessageFormat.format("sending {0} msgs at max inflight window", new Object[]{Integer.valueOf(mActualInFlightCount)}));
                    }

                    throw new MqttException(MqttException.REASON_CODE_MAX_INFLIGHT);
                }

                MqttMessage innerMessage = ((MqttPublish) message).getMessage();
                if (DEBUG) {
                    LogEx.v(LOG_TAG,
                            "[send] "
                                    + MessageFormat.format("pending publish key={0} qos={1} message={2}",
                                    new Object[]{Integer.valueOf(message.getMessageId()), Integer.valueOf(innerMessage.getQos()), message}));
                }

                switch (innerMessage.getQos()) {
                    case 2:
                        mOutboundQoS2Map.put(Integer.valueOf(message.getMessageId()), message);
                        mPersistence.put(getSendPersistenceKey(message), (MqttPublish) message);
                        break;
                    case 1:
                        // Qos 1의 경우 처리.
                        mOutboundQoS1Map.put(Integer.valueOf(message.getMessageId()), message);
                        mPersistence.put(getSendPersistenceKey(message), (MqttPublish) message);
                        break;
                }
                // 요청 추적 위해 token 저장.
                mTokenStore.saveToken(token, message);
                // 메시지 큐에 삽입. / publish 메시지 경우 mPendingMessages에 추가.
                mPendingMessages.addElement(message);
                mQueueLock.notifyAll();
            }
        } else {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[send] " + MessageFormat.format("pending send key={0} message {1}", new Object[]{Integer.valueOf(message.getMessageId()), message}));
            }

            // publish 메시지 아닌 경우 mPendingFlows에 추가.

            //
            // CONNECT와 그외로 나뉜다.
            //

            if (message instanceof MqttConnect) {
                synchronized (mQueueLock) {
                    // Add the connect action at the head of the pending queue ensuring it jumps
                    // ahead of any of other pending actions.
                    mTokenStore.saveToken(token, message);
                    // connect 메시지 경우 mPendingFlowQueue 맨 앞에 추가.
                    mPendingFlowQueue.insertElementAt(message, 0);
                    mQueueLock.notifyAll();
                }
            } else {
                if (message instanceof MqttPingReq) {
                    // PING 요청 시.
                    this.mPingReqMessage = message; // (jaebok.lee) 있을텐데 왜 다시 덮지?
                } else if (message instanceof MqttPubRel) {
                    mOutboundQoS2Map.put(Integer.valueOf(message.getMessageId()), message); // thread-safe
                    mPersistence.put(getSendConfirmPersistenceKey(message), (MqttPubRel) message);
                } else if (message instanceof MqttPubComp) {
                    mPersistence.remove(getReceivedPersistenceKey(message)); // thread-safe
                }

                synchronized (mQueueLock) {
                    if (!(message instanceof MqttAck)) {
                        mTokenStore.saveToken(token, message);
                    }
                    // connect 메시지 경우 mPendingFlowQueue 맨 뒤에 추가.
                    mPendingFlowQueue.addElement(message);
                    mQueueLock.notifyAll();
                }
            }
        }
    }

    /**
     * This removes the MqttSend message from the outbound queue and persistence.
     *
     * @param message
     * @throws MqttPersistenceException
     */
    protected void undo(MqttPublish message) throws MqttPersistenceException {
        synchronized (mQueueLock) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[undo] " + MessageFormat.format("key={0} QoS={1}", new Object[]{Integer.valueOf(message.getMessageId()), Integer.valueOf(message.getMessage().getQos())}));
            }

            if (message.getMessage().getQos() == 1) {
                mOutboundQoS1Map.remove(Integer.valueOf(message.getMessageId()));
            } else {
                mOutboundQoS2Map.remove(Integer.valueOf(message.getMessageId()));
            }
            mPendingMessages.removeElement(message);
            mPersistence.remove(getSendPersistenceKey(message));
            mTokenStore.removeToken(message);
            checkQuiesceLock();
        }
    }

    /**
     * 타이머 쓰레드에서 호출.
     * Check and send a ping if needed and check for ping timeout.
     * Need to send a ping if nothing has been sent or received
     * in the last keepalive interval. It is important to check for
     * both sent and received packets in order to catch the case where an
     * app is solely sending QoS 0 messages or receiving QoS 0 messages.
     * QoS 0 message are not good enough for checking a connection is
     * alive as they are one way messages.
     * <p/>
     * If a ping has been sent but no data has been received in the last keepalive interval then the connection is deamed to be broken.
     * PINGREQ가 전송되었다면 해당 Token 리턴. 아니면 null.
     *
     * @return token of ping command, null if no ping command has been sent.
     */
    public MqttToken checkForActivity() throws MqttException {
        synchronized (mQuiesceLock) {
            // ref bug: https://bugs.eclipse.org/bugs/show_bug.cgi?id=440698
            if (mQuiescing) {
                // 정리 중이라면 처리 안함. 다음 호출을 처리하지 않으므로 (사용자가 직접 호출하지 않는 이상) 이후로 다시 불리우지 않음.
                return null;
            }
        }

        MqttToken token = null;
        long nextPingTime = mKeepAliveInMillis;

        // (jaebok.lee) mConnected는 lock 보호 안 받아도 될까?
        if (mConnected && mKeepAliveInMillis > 0) {
            long timeInMillis = System.currentTimeMillis();
            //Reduce schedule frequency since System.currentTimeMillis is no accurate, add a buffer
            //It is 1/10 in minimum keepalive unit.
            int delta = 100;

            synchronized (mPingLock) {

                if (mIsPingOutstanding.booleanValue() == false) {
                    // 나가 있는 PINGREQ 메시지가 없는 경우.

                    //This follow previous Paho code logic. Should we change to the latest activity time? which is
                    //long lastActivity = lastInboundActivity > lastOutboundActivity ? lastInboundActivity: lastOutboundActivity;
                    //So we delay the ping a bit more. In most cases, lastInboundActivity and lastOutboundActivity has a TTL difference.
                    long lastActivity = mLastInboundActivityInMillis > mLastOutboundActivityTimeInMillis ? mLastOutboundActivityTimeInMillis : mLastInboundActivityInMillis;

                    // Is a ping required?
                    if (timeInMillis - lastActivity >= mKeepAliveInMillis) {

                        if (DEBUG) {
                            LogEx.v(LOG_TAG,
                                    "[checkForActivity] "
                                            + MessageFormat.format("ping needed. keepAlive={0} lastOutboundActivity={1} lastInboundActivity={2}", new Object[]{Long.valueOf(this.mKeepAliveInMillis),
                                            Long.valueOf(mLastOutboundActivityTimeInMillis), Long.valueOf(mLastInboundActivityInMillis)}));
                        }
                        mIsPingOutstanding = Boolean.TRUE;
                        mLastPingTimeInMillis = timeInMillis;
                        token = new MqttToken();
                        mTokenStore.saveToken(token, mPingReqMessage);
                        mPendingFlowQueue.insertElementAt(mPingReqMessage, 0);

                        nextPingTime = Math.min(mKeepAliveInMillis, PING_TIMEOUT_IN_MILLS);

                        //Wake sender thread since it may be in wait state (in ClientState.get())
                        notifyQueueLock();
                    } else {
                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[checkForActivity] " + "ping not needed yet. Schedule next ping.");
                        }
                        nextPingTime = mKeepAliveInMillis - (timeInMillis - lastActivity);
                    }
                } else {
                    // PINGREQ 메시지 보낸 경우.
                    if (/*mClientComms.mCommsReceiver.isReceiving() == false && <- receive에서 멈춰버리는 현상 있음.*/
                            timeInMillis - mLastPingTimeInMillis >= PING_TIMEOUT_IN_MILLS + delta &&
                                    timeInMillis - mLastInboundActivityInMillis >= PING_TIMEOUT_IN_MILLS + delta &&
                                    timeInMillis - mLastOutboundActivityTimeInMillis >= PING_TIMEOUT_IN_MILLS + delta) {
                        // 시간이 지나도록 PINGRESP 가 오지 않은 경우.

                        // any of the conditions is true means the client is active
                        // lastInboundActivity will be updated once receiving is done.
                        //Add a delta, since the timer and System.currentTimeMillis() is not accurate.
                        // A ping is outstanding but no packet has been received in KA so connection is deemed broken
                        if (LogExUtils.isLogExError()) {
                            LogEx.e(LOG_TAG, "[checkForActivity] " + MessageFormat.format("Timed out as no activity, keepAlive={0} lastOutboundActivity={1} lastInboundActivity={2} time={3} lastPing={4}",
                                    new Object[]{Long.valueOf(this.mKeepAliveInMillis), Long.valueOf(mLastOutboundActivityTimeInMillis), Long.valueOf(mLastInboundActivityInMillis), Long.valueOf(timeInMillis), Long.valueOf(mLastPingTimeInMillis)}));
                        }

                        // A ping has already been sent. At this point, assume that the
                        // broker has hung and the TCP layer hasn't noticed.
                        throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_CLIENT_TIMEOUT);
                    } else {
                        nextPingTime = Math.min(mKeepAliveInMillis, PING_TIMEOUT_IN_MILLS);
                    }
                }
            }
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[checkForActivity] " + MessageFormat.format("Schedule next ping at {0}",
                        new Object[]{Long.valueOf(nextPingTime)}));
            }

            mPingSender.schedule(nextPingTime);
        }

        return token;
    }

    /**
     * CommsSender 쓰레드에서 메시지를 얻어감.
     * This returns the next piece of work, ie message, for the CommsSender
     * to send over the network.
     * Calls to this method block until either:
     * - there is a message to be sent
     * - the keepAlive interval is exceeded, which triggers a ping message
     * to be returned
     * - {@link #disconnected(MqttException)} is called
     *
     * @return the next message to send, or null if the client is disconnected
     */
    protected MqttWireMessage get() throws MqttException {
        MqttWireMessage result = null;

        synchronized (mQueueLock) {
            while (result == null) {

                // If there is no work wait until there is work.
                // If the inflight window is full and no flows are pending wait until space is freed.
                // In both cases mQueueLock will be notified.
                if (mPendingMessages.isEmpty() && mPendingFlowQueue.isEmpty() ||
                        mPendingFlowQueue.isEmpty() && mActualInFlightCount >= MAX_INFLIGHT) {
                    try {
                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[get] " + "wait for new work or for space in the inflight window");
                        }

                        mQueueLock.wait();

                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[get] " + "new work or ping arrived");
                        }
                    } catch (InterruptedException ignored) {
                    }
                }

                // Handle the case where not connected. This should only be the case if:
                // - in the process of disconnecting / shutting down
                // - in the process of connecting
                if (!mConnected &&
                        (mPendingFlowQueue.isEmpty() || !(mPendingFlowQueue.elementAt(0) instanceof MqttConnect))) {
                    if (DEBUG) {
                        LogEx.v(LOG_TAG, "[get] " + "no outstanding flows and not connected");
                    }

                    return null;
                }

                // Check if there is a need to send a ping to keep the session alive.
                // Note this check is done before processing messages. If not done first
                // an app that only publishes QoS 0 messages will prevent keepalive processing
                // from functioning.
//                checkForActivity(); //Use pinger, don't check here

                // Now process any queued flows or messages
                if (!mPendingFlowQueue.isEmpty()) {
                    // Process the first "flow" in the queue
                    result = mPendingFlowQueue.remove(0);
                    if (result instanceof MqttPubRel) {
                        mInFlightPubRelCount++;

                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[get] " + MessageFormat.format("+1 inflightpubrels={0}", new Object[]{Integer.valueOf(mInFlightPubRelCount)}));
                        }
                    }

                    checkQuiesceLock();
                } else if (!mPendingMessages.isEmpty()) {
                    // If the inflight window is full then messages are not
                    // processed until the inflight window has space.
                    if (mActualInFlightCount < MAX_INFLIGHT) {
                        // The in flight window is not full so process the
                        // first message in the queue
                        result = mPendingMessages.elementAt(0);
                        mPendingMessages.removeElementAt(0);
                        mActualInFlightCount++;

                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[get] " + MessageFormat.format("+1 actualInFlight={0}", new Object[]{Integer.valueOf(mActualInFlightCount)}));
                        }
                    } else {
                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[get] " + "inflight window full");
                        }
                    }
                }
            }
        }
        return result;
    }

    public void setKeepAliveInterval(long interval) {
        this.mKeepAliveInMillis = interval;
    }

    /**
     * Deduce how long to to wait until a ping is required.
     * <p/>
     * In order to keep the connection alive the server must see activity within the keepalive interval. If the application is not sending / receiving any messages then the client will send a ping.
     * This method works out the next time that a ping must be sent in order for the server to know the client is alive.
     *
     * @return time before a ping needs to be sent to keep alive the connection
     */
    long getTimeUntilPing() {
        long pingin = mKeepAliveInMillis;
        // If KA is zero which means just wait for work or
        // if a ping is outstanding return the KA value
        if (mConnected && mKeepAliveInMillis > 0 && !mIsPingOutstanding.booleanValue()) {

            long time = System.currentTimeMillis();
            long timeSinceOut = time - mLastOutboundActivityTimeInMillis;
            long timeSinceIn = time - mLastInboundActivityInMillis;

            if (timeSinceOut > timeSinceIn) {
                pingin = mKeepAliveInMillis - timeSinceOut;
            } else {
                pingin = mKeepAliveInMillis - timeSinceIn;
            }

            // Unlikely to be negative or zero but in the case it is return a
            // small value > 0 to cause a ping to occur
            if (pingin <= 0) {
                pingin = 10;
            }
        }
        return pingin;
    }

    /**
     * CommsSender 쓰레드로부터 호출됨.
     * Called by the CommsSender when a message has been sent
     *
     * @param message
     */
    protected void notifySent(MqttWireMessage message) {
        this.mLastOutboundActivityTimeInMillis = System.currentTimeMillis();
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[notifySent] " + MessageFormat.format("key={0}", new Object[]{message.getKey()}));
        }

        MqttToken token = mTokenStore.getToken(message);
        token.internalTok.notifySent();
        if (message instanceof MqttPublish) {
            if (((MqttPublish) message).getMessage().getQos() == 0) {
                // Qos 0의 경우에는 보내는 것으로 완료.
                // once a QoS 0 message is sent we can clean up its records straight away as
                // we won't be hearing about it again
                token.internalTok.markComplete(null, null);
                mCallbackThread.asyncOperationComplete(token);
                decrementInFlight();
                releaseMessageId(message.getMessageId());
                mTokenStore.removeToken(message);
                checkQuiesceLock();
            }
        }
    }

    private void decrementInFlight() {
        synchronized (mQueueLock) {
            mActualInFlightCount--;
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[decrementInFlight] " + MessageFormat.format("-1 actualInFlight={0}", new Object[]{Integer.valueOf(mActualInFlightCount)}));
            }

            if (!checkQuiesceLock()) {
                mQueueLock.notifyAll();
            }
        }
    }

    /**
     * 작업 중단 기능을 위한 체크.
     *
     * @return
     */
    protected boolean checkQuiesceLock() {
        int tokenCount = mTokenStore.count();
        if (mQuiescing && tokenCount == 0 && mPendingFlowQueue.size() == 0 && mCallbackThread.isQuiesced()) {
            if (DEBUG) {
                LogEx.v(LOG_TAG,
                        "[checkQuiesceLock] "
                                + MessageFormat.format("mQuiescing={0} mActualInFlightCount={1} mPendingFlowQueue={2} mInFlightPubRelCount={3} callbackQuiesce={4} tokenCount={5}",
                                new Object[]{Boolean.valueOf(mQuiescing), Integer.valueOf(mActualInFlightCount), Integer.valueOf(mPendingFlowQueue.size()), Integer.valueOf(mInFlightPubRelCount),
                                        Boolean.valueOf(mCallbackThread.isQuiesced()), Integer.valueOf(tokenCount)}));
            }
            synchronized (mQuiesceLock) {
                mQuiesceLock.notifyAll();
            }
            return true;
        }
        return false;
    }

    /**
     * Called by the CommsReceiver when an ack has arrived.
     * MqttAck 종류 수신의 경우 호출됨.
     * CommsReceiver 쓰레드에서 동작.
     *
     * @param ack ack message
     * @throws MqttException
     */
    protected void notifyReceivedAck(MqttAck ack) throws MqttException {
        this.mLastInboundActivityInMillis = System.currentTimeMillis();

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[notifyReceivedAck] " + MessageFormat.format("received key={0} message={1}",
                    new Object[]{Integer.valueOf(ack.getMessageId()), ack}));
        }

        MqttToken token = mTokenStore.getToken(ack);

        if (ack instanceof MqttPubRec) {
            // Complete the QoS 2 flow. Unlike all other
            // flows, QoS is a 2 phase flow. The second phase sends a
            // PUBREL - the operation is not complete until a PUBCOMP
            // is received
            MqttPubRel rel = new MqttPubRel((MqttPubRec) ack);
            this.send(rel, token);
        } else if (ack instanceof MqttPubAck || ack instanceof MqttPubComp) {
            // QoS 1 & 2 notify users of result before removing from
            // persistence
            notifyResult(ack, token, null);
            // Do not remove publish / delivery token at this stage
            // do this when the persistence is removed later
        } else if (ack instanceof MqttPingResp) {
            // PING 응답 경우.
            synchronized (mPingLock) {
                mIsPingOutstanding = Boolean.FALSE; // 핑 응답 받음.
                notifyResult(ack, token, null);
                mTokenStore.removeToken(ack);
            }
        } else if (ack instanceof MqttConnack) {
            // CONNECT 응답 경우.
            int rc = ((MqttConnack) ack).getReturnCode();
            if (rc == 0) {
                synchronized (mQueueLock) {
                    if (mCleanSession) {
                        clearState();
                        // Add the connect token back in so that users can be
                        // notified when connect completes.
                        // 위 메소드에서 mTokenStore clear됨.
                        mTokenStore.saveToken(token, ack);
                    }
                    mInFlightPubRelCount = 0;
                    mActualInFlightCount = 0;
                    restoreInflightMessages(); // CleanSession true 경우 무의미.
                    connected(); // ping sender 시작.
                }
            } else {
                // (jaebok.lee) 인증이 만료되어 실패할 경우 이 경로로 들어온다.
                throw ExceptionHelper.createMqttException(rc);
            }

            // 연결 성공 알림. mex == null
            mClientComms.connectComplete((MqttConnack) ack);
            notifyResult(ack, token, null);
            mTokenStore.removeToken(ack);

            // Notify the sender thread that there maybe work for it to do now
            synchronized (mQueueLock) {
                mQueueLock.notifyAll();
            }
        } else {
            // Sub ack or unsuback
            notifyResult(ack, token, null);
            releaseMessageId(ack.getMessageId());
            mTokenStore.removeToken(ack);
        }

        checkQuiesceLock();
    }

    /**
     * CommsReceiver 쓰레드로부터 MqttAck 종류가 아닌 경우 호출.
     * Called by the CommsReceiver when a message has been received.
     * Handles inbound messages and other flows such as PUBREL.
     *
     * @param message
     * @throws MqttException
     */
    protected void notifyReceivedMsg(MqttWireMessage message) throws MqttException {
        this.mLastInboundActivityInMillis = System.currentTimeMillis();

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[notifyReceivedMsg] " + MessageFormat.format("received key={0} message={1}",
                    new Object[]{Integer.valueOf(message.getMessageId()), message}));
        }

        if (!mQuiescing) {
            if (message instanceof MqttPublish) {
                MqttPublish send = (MqttPublish) message;
                switch (send.getMessage().getQos()) {
                    case 0:
                    case 1:
                        if (mCallbackThread != null) {
                            mCallbackThread.messageArrived(send);
                        }
                        break;
                    case 2:
                        mPersistence.put(getReceivedPersistenceKey(message),
                                (MqttPublish) message);
                        mInboundQoS2Map.put(Integer.valueOf(send.getMessageId()), send);
                        this.send(new MqttPubRec(send), null);
                }
            } else if (message instanceof MqttPubRel) {
                MqttPublish sendMsg = (MqttPublish) mInboundQoS2Map
                        .get(Integer.valueOf(message.getMessageId()));
                if (sendMsg != null) {
                    if (mCallbackThread != null) {
                        mCallbackThread.messageArrived(sendMsg);
                    }
                } else {
                    // Original publish has already been delivered.
                    MqttPubComp pubComp = new MqttPubComp(message
                            .getMessageId());
                    this.send(pubComp, null);
                }
            }
        }
    }

    /**
     * MqttAck 관련 메시지 처리.
     * Called when waiters and callbacks have processed the message. For
     * messages where delivery is complete the message can be removed from
     * persistence and counters adjusted accordingly. Also tidy up by removing
     * token from store...
     *
     * @param token
     * @throws MqttException
     */
    protected void notifyComplete(MqttToken token) throws MqttException {
        MqttWireMessage message = token.internalTok.getWireMessage();

        if (message != null && message instanceof MqttAck) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[notifyComplete] " + MessageFormat.format("received key={0} token={1} message={2}",
                        new Object[]{Integer.valueOf(message.getMessageId()), token, message}));
            }

            MqttAck ack = (MqttAck) message;

            if (ack instanceof MqttPubAck) {
                // QoS 1 경우 - user notified now remove from persistence...
                mPersistence.remove(getSendPersistenceKey(message));
                mOutboundQoS1Map.remove(Integer.valueOf(ack.getMessageId()));
                decrementInFlight();
                releaseMessageId(message.getMessageId()); // message id 회수.
                mTokenStore.removeToken(message);
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[notifyComplete] " + MessageFormat.format("removed Qos 1 publish. key={0}",
                            new Object[]{Integer.valueOf(ack.getMessageId())}));
                }
            } else if (ack instanceof MqttPubComp) {
                // QoS 2 - user notified now remove from persistence...
                mPersistence.remove(getSendPersistenceKey(message));
                mPersistence.remove(getSendConfirmPersistenceKey(message));
                mOutboundQoS2Map.remove(Integer.valueOf(ack.getMessageId()));

                mInFlightPubRelCount--;
                decrementInFlight();
                releaseMessageId(message.getMessageId());
                mTokenStore.removeToken(message);

                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[notifyComplete] " + MessageFormat.format("removed QoS 2 publish/pubrel. key={0}, -1 inFlightPubRels={1}",
                            new Object[]{Integer.valueOf(ack.getMessageId()), Integer.valueOf(mInFlightPubRelCount)}));
                }
            }

            checkQuiesceLock();
        }
    }

    /**
     * 응답 알림 이벤트.
     * CommsReceiver 쓰레드에서 동작.
     *
     * @param ack
     * @param token
     * @param ex
     */
    protected void notifyResult(MqttAck ack, MqttToken token, MqttException ex) {
        // unblock any threads waiting on the token
        token.internalTok.markComplete(ack, ex);

        // Let the user know an async operation has completed and then remove the token
        if (ack != null && !(ack instanceof MqttPubRec)) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[notifyResult] " + MessageFormat.format("key{0}, msg={1}, excep={2}",
                        new Object[]{token.internalTok.getKey(), ack, ex}));

            }
            mCallbackThread.asyncOperationComplete(token);
        }
        // There are cases where there is no ack as the operation failed before
        // an ack was received
        if (ack == null) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[notifyResult] " + MessageFormat.format("key={0},excep={1}",
                        new Object[]{token.internalTok.getKey(), ex}));
            }
            mCallbackThread.asyncOperationComplete(token);
        }
    }

    /**
     * Called when the client has successfully connected to the broker
     * Ping을 보내기 시작한다.
     */
    public void connected() {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[connected] " + "connected");

        }
        this.mConnected = true;

        mPingSender.start(); //Start ping thread when client connected to server.
    }

    /**
     * Called during shutdown to work out if there are any tokens still
     * to be notified and waiters to be unblocked. Notifying and unblocking
     * takes place after most shutdown processing has completed. The tokenstore
     * is tidied up so it only contains outstanding delivery tokens which are
     * valid after reconnect (if clean session is false)
     *
     * @param reason The root cause of the disconnection, or null if it is a clean disconnect
     */
    public Vector<MqttToken> resolveOldTokens(MqttException reason) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[resolveOldTokens] " + MessageFormat.format("reason {0}",
                    new Object[]{reason}));
        }

        // If any outstanding let the user know the reason why it is still
        // outstanding by putting the reason shutdown is occurring into the
        // token.
        MqttException shutReason = reason;
        if (reason == null) {
            shutReason = new MqttException(MqttException.REASON_CODE_CLIENT_DISCONNECTING);
        }

        // Set the token up so it is ready to be notified after disconnect
        // processing has completed. Do not
        // remove the token from the store if it is a delivery token, it is
        // valid after a reconnect.
        Vector<MqttToken> outstandingTokens = mTokenStore.getOutstandingTokens();
        Enumeration<MqttToken> outTE = outstandingTokens.elements();
        while (outTE.hasMoreElements()) {
            final MqttToken token = outTE.nextElement();
            synchronized (token) {
                if (token.isComplete() == false && !token.internalTok.isCompletePending() && token.getException() == null) {
                    token.internalTok.setException(shutReason);
                }
            }
            if ((token instanceof MqttDeliveryToken) == false) {
                // If not a delivery token it is not valid on
                // restart so remove
                mTokenStore.removeToken(token.internalTok.getKey());
            }
        }
        return outstandingTokens;
    }

    /**
     * Called when the client has been disconnected from the broker.
     *
     * @param reason The root cause of the disconnection, or null if it is a clean disconnect
     */
    public void disconnected(MqttException reason) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[disconnected] " + "disconnected");
        }

        mConnected = false;

        try {
            if (mCleanSession) {
                clearState();
            }

            mPendingMessages.clear();
            mPendingFlowQueue.clear();
            synchronized (mPingLock) {
                // Reset pingOutstanding to allow reconnects to assume no previous ping.
                mIsPingOutstanding = Boolean.FALSE;
            }
        } catch (MqttException e) {
            // Ignore as we have disconnected at this point
        }
    }

    /**
     * message id 회수.
     * Releases a message ID back into the pool of available message IDs.
     * If the supplied message ID is not in use, then nothing will happen.
     *
     * @param msgId A message ID that can be freed up for re-use.
     */
    private synchronized void releaseMessageId(int msgId) {
        mInUseMsgIdMap.remove(Integer.valueOf(msgId));
    }

    /**
     * Get the next MQTT message ID that is not already in use, and marks
     * it as now being in use.
     *
     * @return the next MQTT message ID to use
     */
    private synchronized int getNextMessageId() throws MqttException {
        int startingMessageId = mNextMsgId;
        // Allow two complete passes of the message ID range. This gives
        // any asynchronous releases a chance to occur
        int loopCount = 0;
        do {
            mNextMsgId++;
            if (mNextMsgId > MAX_MSG_ID) {
                mNextMsgId = MIN_MSG_ID;
            }
            if (mNextMsgId == startingMessageId) {
                loopCount++;
                if (loopCount == 2) {
                    throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_NO_MESSAGE_IDS_AVAILABLE);
                }
            }
        } while (mInUseMsgIdMap.containsKey(Integer.valueOf(mNextMsgId)));
        Integer id = Integer.valueOf(mNextMsgId);
        mInUseMsgIdMap.put(id, id);
        return mNextMsgId;
    }

    /**
     * 작업 중단. 종료 시나리오에서 호출.
     * Quiesce the client state, preventing any new messages getting sent,
     * and preventing the callback on any newly received messages.
     * After the timeout expires, delete any pending messages except for
     * outbound ACKs, and wait for those ACKs to complete.
     */
    public void quiesce(long timeout) {
        // If the timeout is greater than zero t
        if (timeout > 0) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[quiesce] " + MessageFormat.format("timeout={0}", new Object[]{Long.valueOf(timeout)}));
            }
            synchronized (mQueueLock) {
                mQuiescing = true;
            }

            // TODO lock ?
            // We don't want to handle any new inbound messages
            mCallbackThread.quiesce();
            notifyQueueLock();

            synchronized (mQuiesceLock) {
                try {
                    // If token count is not zero there is outbound work to process and
                    // if pending flows is not zero there is outstanding work to complete and
                    // if call back is not quiseced there it needs to complete.
                    int tokenCount = mTokenStore.count();
                    if (tokenCount > 0 || mPendingFlowQueue.size() > 0 || mCallbackThread.isQuiesced() == false) {
                        if (DEBUG) {
                            LogEx.v(LOG_TAG,
                                    "[quiesce] "
                                            + MessageFormat.format("wait for outstanding: mActualInFlightCount={0} mPendingFlowQueue={1} mInFlightPubRelCount={2} tokenCount={3}",
                                            new Object[]{Integer.valueOf(mActualInFlightCount), Integer.valueOf(mPendingFlowQueue.size()), Integer.valueOf(mInFlightPubRelCount), Integer.valueOf(tokenCount)}));
                        }

                        // wait for outstanding in flight messages to complete and
                        // any pending flows to complete
                        mQuiesceLock.wait(timeout);
                    }
                } catch (InterruptedException ignored) {
                    // Don't care, as we're shutting down anyway
                }
            }

            // Quiesce time up or inflight messages delivered.  Ensure pending delivery
            // vectors are cleared ready for disconnect to be sent as the final flow.
            synchronized (mQueueLock) {
                mPendingMessages.clear();
                mPendingFlowQueue.clear();
                mQuiescing = false;
                mActualInFlightCount = 0;
            }
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[quiesce] " + "finished");
            }
        }
    }

    public void notifyQueueLock() {
        synchronized (mQueueLock) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[notifyQueueLock] " + "notifying mQueueLock holders");
            }
            mQueueLock.notifyAll();
        }
    }

    protected void deliveryComplete(MqttPublish message) throws MqttPersistenceException {

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[deliveryComplete] " + MessageFormat.format("remove publish from persistence. key={0}", new Object[]{Integer.valueOf(message.getMessageId())}));
        }

        mPersistence.remove(getReceivedPersistenceKey(message)); // thread-safe
        mInboundQoS2Map.remove(Integer.valueOf(message.getMessageId())); // thread-safe
    }

    /**
     * Tidy up
     * - ensure that tokens are released as they are maintained over a
     * disconnect / connect cycle.
     * ClientComms.mConnectLock 내에서 호출.
     */
    protected void close() {
        mInUseMsgIdMap.clear(); // thread-safe
        mPendingMessages.clear(); // thread-safe
        mPendingFlowQueue.clear(); // thread-safe
        mOutboundQoS2Map.clear(); // thread-safe
        mOutboundQoS1Map.clear(); // thread-safe
        mInboundQoS2Map.clear(); // thread-safe
        mTokenStore.clear(); // thread-safe
        mInUseMsgIdMap = null;
        mPendingMessages = null;
        mPendingFlowQueue = null;
        mOutboundQoS2Map = null;
        mOutboundQoS1Map = null;
        mInboundQoS2Map = null;
        mTokenStore = null;
        mCallbackThread = null;
        mClientComms = null;
        mPersistence = null;
        mPingReqMessage = null;
    }

}
