/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.comm.lib.utils.LogExUtils;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.IMqttActionListener;
import org.eclipse.paho.client.mqttv3.MqttCallback;
import org.eclipse.paho.client.mqttv3.MqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttToken;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPubAck;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPubComp;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPublish;

import java.text.MessageFormat;
import java.util.Vector;

/**
 * Bridge between Receiver and the external API. This class gets called by
 * Receiver, and then converts the comms-centric MQTT message objects into ones
 * understood by the external API.
 */
public class CommsCallback implements Runnable {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "CommsCallback";

    /**
     * PUBLISH 메시지 수신 큐에 대한 제한.
     */
    private static final int INBOUND_QUEUE_SIZE = 10;

    private MqttCallback mMqttCallback;
    private ClientComms mClientComms;

    /**
     * PUBLISH 메시지 수신 큐.
     */
    private final Vector<MqttPublish> mMessageQueue;
    /**
     * 처리 완료 큐
     */
    private final Vector<MqttToken> mCompleteQueue;
    public boolean mIsRunning = false;
    private boolean mQuiescing = false;
    private final Object mLifecycleLock = new Object();
    private Thread mThread;
    /**
     * mMessageQueue, mCompleteQueue를 위한 락.
     */
    private final Object mWorkAvailable = new Object();
    private final Object mSpaceAvailable = new Object();
    private ClientState mClientState;

    CommsCallback(ClientComms clientComms) {
        this.mClientComms = clientComms;
        this.mMessageQueue = new Vector<MqttPublish>(INBOUND_QUEUE_SIZE);
        this.mCompleteQueue = new Vector<MqttToken>(INBOUND_QUEUE_SIZE);
    }

    public void setClientState(ClientState clientState) {
        this.mClientState = clientState;
    }

    /**
     * Starts up the Callback thread.
     */
    public void start(String threadName) {
        synchronized (mLifecycleLock) {
            if (mIsRunning == false) {
                // Praparatory work before starting the background thread.
                // For safety ensure any old events are cleared.
                mMessageQueue.clear();
                mCompleteQueue.clear();

                mIsRunning = true;
                mQuiescing = false;
                mThread = new Thread(this, threadName);
                mThread.start();
            }
        }
    }

    /**
     * Stops the callback thread.
     * This call will block until stop has completed.
     */
    public void stop() {
        synchronized (mLifecycleLock) {
            if (mIsRunning) {
                if (DEBUG) {
                    // @TRACE 700=stopping
                    LogEx.v(LOG_TAG, "[stop] " + "stopping");
                }
                mIsRunning = false;
                if (Thread.currentThread().equals(mThread) == false) {
                    try {
                        synchronized (mWorkAvailable) {
                            if (DEBUG) {
                                // @TRACE 701=notify mWorkAvailable and wait for run
                                // to finish
                                LogEx.v(LOG_TAG, "[stop] " + "notify mWorkAvailable and wait for run");
                            }
                            mWorkAvailable.notifyAll();
                        }
                        // Wait for the thread to finish.
                        mThread.join();
                    } catch (InterruptedException ignored) {
                    }
                }
            }
            mThread = null;
            if (DEBUG) {
                // @TRACE 703=stopped
                LogEx.v(LOG_TAG, "[stop] " + "stopped");
            }
        }
    }

    public void setCallback(MqttCallback mqttCallback) {
        mMqttCallback = mqttCallback;
    }

    @Override
    public void run() {
        while (mIsRunning) {
            try {
                // If no work is currently available, then wait until there is some...
                try {
                    synchronized (mWorkAvailable) {
                        // 큐가 비어 있으면 대기.
                        if (mIsRunning & mMessageQueue.isEmpty()
                                && mCompleteQueue.isEmpty()) {
                            if (DEBUG) {
                                // @TRACE 704=wait for mWorkAvailable
                                LogEx.v(LOG_TAG, "[run] " + "wait for mWorkAvailable");
                            }
                            mWorkAvailable.wait();
                        }
                    }
                } catch (InterruptedException ignored) {
                }

                if (mIsRunning) {
                    // Check for deliveryComplete callbacks...
                    MqttToken token = null;
                    synchronized (mCompleteQueue) {
                        if (!mCompleteQueue.isEmpty()) {
                            // First call the delivery arrived callback if needed
                            token = mCompleteQueue.elementAt(0);
                            mCompleteQueue.removeElementAt(0);
                        }
                    }
                    if (null != token) {
                        // 토큰 완료 처리.
                        handleActionComplete(token);
                    }

                    // Check for messageArrived callbacks...
                    MqttPublish message = null;
                    synchronized (mMessageQueue) {
                        if (!mMessageQueue.isEmpty()) {
                            // Note, there is a window on connect where a publish
                            // could arrive before we've
                            // finished the connect logic.
                            message = mMessageQueue.elementAt(0);

                            mMessageQueue.removeElementAt(0);
                        }
                    }
                    if (null != message) {
                        // 수신 메시지 처리.
                        handleMessage(message);
                    }
                }

                if (mQuiescing) {
                    mClientState.checkQuiesceLock();
                }

                synchronized (mSpaceAvailable) {
                    // Notify the spaceAvailable lock, to say that there's now
                    // some space on the queue...

                    if (DEBUG) {
                        // @TRACE 706=notify spaceAvailable
                        LogEx.v(LOG_TAG, "[run] " + "notify spaceAvailable");
                    }
                    mSpaceAvailable.notifyAll();
                }
            } catch (Throwable ex) {
                // Users code could throw an Error or Exception e.g. in the case
                // of class NoClassDefFoundError
                if (DEBUG) {
                    // @TRACE 714=callback threw exception
                    LogEx.v(LOG_TAG, "[run] " + "callback threw exception");
                }
                mIsRunning = false;
                mClientComms.shutdownConnection(null, new MqttException(ex));
            }
        }
    }

    /**
     * CommsCallback 쓰레드에서 호출.
     * 쓰레드 미동작시에는 CommsReceiver / DisconnectBG 쓰레드에서 동작할 수도 있음.
     *
     * @param token
     * @throws MqttException
     */
    private void handleActionComplete(MqttToken token) throws MqttException {
        synchronized (token) {
            if (DEBUG) {
                // @TRACE 705=callback and notify for key={0}
                LogEx.v(LOG_TAG, "[handleActionComplete] " + MessageFormat.format("callback and notify for key={0}", new Object[]{token.internalTok.getKey()}));
            }

            // block 해제.
            // Unblock any waiters and if pending complete now set completed
            token.internalTok.notifyComplete();

            if (token.internalTok.isNotified() == false) {
                // If a callback is registered and delivery has finished 
                // call delivery complete callback. 
                if (mMqttCallback != null
                        && token instanceof MqttDeliveryToken
                        && token.isComplete()) {
                    mMqttCallback.deliveryComplete((MqttDeliveryToken) token);
                }
                // Now call async action completion callbacks
                fireActionEvent(token);
            }

            // Set notified so we don't tell the user again about this action.
            if (token instanceof MqttDeliveryToken && token.isComplete()) {
                token.internalTok.setNotified(true);
            }

            if (token.isComplete()) {
                // Finish by doing any post processing such as delete 
                // from persistent store but only do so if the action
                // is complete
                mClientState.notifyComplete(token);
            }
        }
    }

    /**
     * This method is called when the connection to the server is lost. If there
     * is no cause then it was a clean disconnect. The connectionLost callback
     * will be invoked if registered and run on the thread that requested
     * shutdown e.g. receiver or sender thread. If the request was a user
     * initiated disconnect then the disconnect token will be notified.
     *
     * @param cause the reason behind the loss of connection.
     */
    public void connectionLost(MqttException cause) {
        // If there was a problem and a client callback has been set inform
        // the connection lost listener of the problem.
        try {
            if (mMqttCallback != null && cause != null) {
                if (DEBUG) {
                    // @TRACE 708=call connectionLost
                    LogEx.v(LOG_TAG, "[connectionLost] " + "call connectionLost getReasonCode :  " + cause.getReasonCode());
                }
                mMqttCallback.connectionLost(cause);
            }
        } catch (Throwable t) {
            // Just log the fact that a throwable has caught connection lost 
            // is called during shutdown processing so no need to do anything else
            if (DEBUG) {
                // @TRACE 720=exception from connectionLost {0}
                LogEx.v(LOG_TAG, "[connectionLost] " + MessageFormat.format("exception from connectionLost {0}", new Object[]{t}));
            }
        }
    }

    /**
     * An action has completed - if a completion listener has been set on the
     * token then invoke it with the outcome of the action.
     *
     * @param token
     */
    public void fireActionEvent(MqttToken token) {

        if (token != null) {
            IMqttActionListener asyncCB = token.getActionCallback();
            if (asyncCB != null) {
                if (token.getException() == null) {
                    if (DEBUG) {
                        // @TRACE 716=call onSuccess key={0}
                        LogEx.v(LOG_TAG, "[fireActionEvent] " + MessageFormat.format("call onSuccess key={0}", new Object[]{token.internalTok.getKey()}));
                    }
                    asyncCB.onSuccess(token);
                } else {
                    if (DEBUG) {
                        // @TRACE 717=call onFailure key {0}
                        LogEx.v(LOG_TAG, "[fireActionEvent] " + MessageFormat.format("call onFailure key {0}", new Object[]{token.internalTok.getKey()}));
                    }
                    asyncCB.onFailure(token, token.getException());
                }
            }
        }
    }

    /**
     * This method is called when a message arrives on a topic. Messages are
     * only added to the queue for inbound messages if the client is not
     * quiescing.
     *
     * @param sendMessage the MQTT SEND message.
     */
    public void messageArrived(MqttPublish sendMessage) {
        if (mMqttCallback != null) {
            // If we already have enough messages queued up in memory, wait
            // until some more queue space becomes available. This helps 
            // the client protect itself from getting flooded by messages 
            // from the server.
            synchronized (mSpaceAvailable) {
                while (!mQuiescing && mMessageQueue.size() >= INBOUND_QUEUE_SIZE) {
                    try {
                        if (DEBUG) {
                            // @TRACE 709=wait for spaceAvailable
                            LogEx.v(LOG_TAG, "[messageArrived] " + "wait for spaceAvailable");
                        }
                        mSpaceAvailable.wait();
                    } catch (InterruptedException ignored) {
                    }
                }
            }
            if (!mQuiescing) {
                mMessageQueue.addElement(sendMessage);
                // Notify the CommsCallback thread that there's work to do...
                synchronized (mWorkAvailable) {
                    if (DEBUG) {
                        // @TRACE 710=new msg avail, notify mWorkAvailable
                        LogEx.v(LOG_TAG, "[messageArrived] " + "new msg avail, notify mWorkAvailable");
                    }
                    mWorkAvailable.notifyAll();
                }
            }
        }
    }

    /**
     * 작업 중단 요청.
     * Let the call back thread quiesce. Prevent new inbound messages being
     * added to the process queue and let existing work quiesce. (until the
     * thread is told to shutdown).
     */
    public void quiesce() {
        this.mQuiescing = true;
        synchronized (mSpaceAvailable) {
            if (DEBUG) {
                // @TRACE 711=quiesce notify spaceAvailable
                LogEx.v(LOG_TAG, "[quiesce] " + "quiesce notify spaceAvailable");
            }
            // Unblock anything waiting for space...
            mSpaceAvailable.notifyAll();
        }
    }

    /**
     * 작업이 없는지 여부.
     *
     * @return
     */
    public boolean isQuiesced() {
        if (mQuiescing && mCompleteQueue.size() == 0 && mMessageQueue.size() == 0) {
            return true;
        }
        return false;
    }

    private void handleMessage(MqttPublish publishMessage)
            throws MqttException, Exception {
        // If quisecing process any pending messages.
        if (mMqttCallback != null) {
            String destName = publishMessage.getTopicName();

            if (DEBUG) {
                // @TRACE 713=call messageArrived key={0} topic={1}
                LogEx.v(LOG_TAG, "[handleMessage] " + MessageFormat.format("call messageArrived key={0} topic={1}", new Object[]{
                        publishMessage.getMessageId(), destName}));
            }
            mMqttCallback.messageArrived(destName, publishMessage.getMessage());
            if (publishMessage.getMessage().getQos() == 1) {
                // Qos 1의 경우 PUBACK를 전송한다.
                mClientComms.internalSend(new MqttPubAck(publishMessage), new MqttToken());
            } else if (publishMessage.getMessage().getQos() == 2) {
                mClientComms.deliveryComplete(publishMessage);
                MqttPubComp pubComp = new MqttPubComp(publishMessage);
                mClientComms.internalSend(pubComp, new MqttToken());
            }
        }
    }

    /**
     * 토큰 완료? 알림.
     * CommsReceiver 쓰레드에서 동작. 또는 DisconnectBG에서 shutdown 시 호출.
     *
     * @param token
     */
    public void asyncOperationComplete(MqttToken token) {
        if (mIsRunning) {
            // invoke callbacks on callback thread
            mCompleteQueue.addElement(token);
            synchronized (mWorkAvailable) {
                if (DEBUG) {
                    // @TRACE 715=new mWorkAvailable. key={0}
                    LogEx.v(LOG_TAG, "[asyncOperationComplete] " + MessageFormat.format("new mWorkAvailable. key={0}", new Object[]{token.internalTok.getKey()}));
                }
                mWorkAvailable.notifyAll();
            }
        } else {
            // disconnect 경우 이 경로를 타게 됨.
            // invoke async callback on invokers thread
            try {
                handleActionComplete(token);
            } catch (Throwable ex) {
                // Users code could throw an Error or Exception e.g. in the case
                // of class NoClassDefFoundError
                // @TRACE 719=callback threw ex:
                if (LogExUtils.isLogExError()) {
                    LogEx.e(LOG_TAG, "[asyncOperationComplete] " + "callback threw ex:", ex);
                }

                // Shutdown likely already in progress but no harm to confirm
                mClientComms.shutdownConnection(null, new MqttException(ex));
            }

        }
    }

    /**
     * Returns the thread used by this callback.
     */
    protected Thread getThread() {
        return mThread;
    }
}
