/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.comm.lib.utils.LogExUtils;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttToken;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttAck;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttInputStream;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttWireMessage;

import java.io.IOException;
import java.io.InputStream;

/**
 * Receives MQTT packets from the server.
 */
public class CommsReceiver implements Runnable {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "CommsReceiver";

    private boolean mIsRunning = false;
    private final Object mLifecycleLock = new Object();
    private final ClientState mClientState;
    private final ClientComms mClientComms;
    private final MqttInputStream mIn;
    private final CommsTokenStore mTokenStore;
    private Thread mThread = null;
    private volatile boolean mIsReceiving;

    public CommsReceiver(ClientComms clientComms, ClientState clientState, CommsTokenStore tokenStore, InputStream in) {
        this.mIn = new MqttInputStream(in);
        this.mClientComms = clientComms;
        this.mClientState = clientState;
        this.mTokenStore = tokenStore;
    }

    /**
     * Starts up the Receiver's thread.
     */
    public void start(String threadName) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[start] " + "starting");
        }
        synchronized (mLifecycleLock) {
            if (mIsRunning == false) {
                mIsRunning = true;
                mThread = new Thread(this, threadName);
                mThread.start();
            }
        }
    }

    /**
     * Stops the Receiver's thread. This call will block.
     */
    public void stop() {
        synchronized (mLifecycleLock) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[stop] " + "stopping");
            }
            if (mIsRunning) {
                mIsRunning = false;
                mIsReceiving = false;
                if (Thread.currentThread().equals(mThread) == false) {
                    try {
                        // Wait for the thread to finish.
                        mThread.join();
                    } catch (InterruptedException ignored) {
                    }
                }
            }
        }
        mThread = null;
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[stop] " + "stopped");
        }
    }

    /**
     * Run loop to receive messages from the server.
     */
    @Override
    public void run() {
        MqttToken token = null;

        while (mIsRunning && mIn != null) {
            try {
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[run] " + "network read message");
                }
                mIsReceiving = true;
                // 메시지를 읽어옴.
                MqttWireMessage message = mIn.readMqttWireMessage();
                mIsReceiving = false;

                if (message instanceof MqttAck) {
                    // ACK 경우.
                    token = mTokenStore.getToken(message);
                    if (token != null) {
                        // 토큰이 있다면.
                        synchronized (token) {
                            // Ensure the notify processing is done under a lock on the token
                            // This ensures that the send processing can complete  before the 
                            // receive processing starts! ( request and ack and ack processing
                            // can occur before request processing is complete if not!
                            mClientState.notifyReceivedAck((MqttAck) message);
                        }
                    } else {
                        // It its an ack and there is no token then something is not right.
                        // An ack should always have a token assoicated with it.
                        throw new MqttException(MqttException.REASON_CODE_UNEXPECTED_ERROR);
                    }
                } else {
                    // A new message has arrived
                    // 메시지가 왔다면.
                    mClientState.notifyReceivedMsg(message);
                }
            } catch (MqttException ex) {
                // (jaebok.lee) 인증이 만료되어 실패할 경우 이 경로로 들어온다.
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[run] " + "Stopping, MQttException");
                }
                mIsRunning = false;
                // Token maybe null but that is handled mIn shutdown
                mClientComms.shutdownConnection(token, ex);
            } catch (IOException ioe) {
                if (LogExUtils.isLogExWarn()) {
                    LogEx.w(LOG_TAG, "[run] " + "Stopping due to IOException", ioe);
                }

                mIsRunning = false;
                // An EOFException could be raised if the broker processes the 
                // DISCONNECT and ends the socket before we complete. As such,
                // only shutdown the connection if we're not already shutting down.
                if (mClientComms.isDisconnecting() == false && mClientComms.isDisconnectSent() == false) {
                    mClientComms.shutdownConnection(token, new MqttException(MqttException.REASON_CODE_CONNECTION_LOST, ioe));
                }
            } finally {
                mIsReceiving = false;

                // (jaebok.lee) 초기화.
                token = null;
            }
        } // end of while.

        if (DEBUG) {
            LogEx.v(LOG_TAG, "[run] " + "<");
        }
    }

    public boolean isRunning() {
        return mIsRunning;
    }

    /**
     * Returns the mIsReceiving state.
     *
     * @return true if the receiver is mIsReceiving data, false otherwise.
     */
    public boolean isReceiving() {
        return mIsReceiving;
    }
}
