/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.comm.lib.utils.LogExUtils;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttToken;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttAck;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttDisconnect;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttOutputStream;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttWireMessage;

import java.io.IOException;
import java.io.OutputStream;
import java.text.MessageFormat;

public class CommsSender implements Runnable {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "CommsSender";

    /**
     * Sends MQTT packets to the server on its own thread
     */
    private boolean mIsRunning = false;
    private final Object mLifecycleLock = new Object();
    private ClientState mClientState = null;
    private MqttOutputStream mOutputStream;
    private ClientComms mClientComms = null;

    /**
     * ClientState로부터 받은 토큰 저장소.
     */
    private CommsTokenStore mTokenStore = null;
    private Thread mSendThread = null;

    public CommsSender(ClientComms clientComms, ClientState clientState, CommsTokenStore tokenStore, OutputStream out) {
        this.mOutputStream = new MqttOutputStream(out);
        this.mClientComms = clientComms;
        this.mClientState = clientState;
        this.mTokenStore = tokenStore;
    }

    /**
     * Starts up the Sender thread.
     */
    public void start(String threadName) {
        synchronized (mLifecycleLock) {
            if (mIsRunning == false) {
                mIsRunning = true;
                mSendThread = new Thread(this, threadName);
                mSendThread.start();
            }
        }
    }

    /**
     * Stops the Sender's thread.  This call will block.
     */
    public void stop() {
        synchronized (mLifecycleLock) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[stop] " + "stopping sender");
            }
            if (mIsRunning) {
                mIsRunning = false;
                if (!Thread.currentThread().equals(mSendThread)) {
                    try {
                        // first notify get routine to finish
                        mClientState.notifyQueueLock();
                        // Wait for the thread to finish.
                        mSendThread.join();
                    } catch (InterruptedException ignored) {
                    }
                }
            }
            mSendThread = null;
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[stop] " + "stopped");
            }
        }
    }

    public void run() {
        MqttWireMessage message = null;
        while (mIsRunning && (mOutputStream != null)) {
            try {
                message = mClientState.get();
                if (message != null) {
                    if (DEBUG) {
                        LogEx.v(LOG_TAG, "[run] " + MessageFormat.format("network send key={0} msg={1}", new Object[]{message.getKey(), message}));
                    }
                    if (message instanceof MqttAck) {
                        // ACK 메시지 경우.
                        mOutputStream.write(message);
                        mOutputStream.flush();
                    } else {
                        MqttToken token = mTokenStore.getToken(message);
                        // While quiescing the tokenstore can be cleared so need
                        // to check for null for the case where clear occurs
                        // while trying to send a message.
                        if (token != null) {
                            synchronized (token) {
                                mOutputStream.write(message);
                                try {
                                    mOutputStream.flush();
                                } catch (IOException ex) {
                                    if (LogExUtils.isLogExWarn()) {
                                        LogEx.w(LOG_TAG, "[run]", ex);
                                    }

                                    // The flush has been seen to fail on disconnect of a SSL socket
                                    // as disconnect is in progress this should not be treated as an error
                                    if (message instanceof MqttDisconnect == false) {
                                        throw ex;
                                    }
                                }
                                mClientState.notifySent(message);
                            }
                        }
                    }
                } else { // null message
                    if (DEBUG) {
                        LogEx.v(LOG_TAG, "[run] " + "get message returned null, stopping}");
                    }
                    mIsRunning = false;
                }
            } catch (MqttException me) {
                handleRunException(message, me);
            } catch (Exception ex) {
                handleRunException(message, ex);
            }
        } // end while
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[run] " + "<");
        }
    }

    /**
     * 종료 처리.
     *
     * @param message
     * @param ex
     */
    private void handleRunException(MqttWireMessage message, Exception ex) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[handleRunException] " + "exception");
        }
        MqttException mex;
        if (!(ex instanceof MqttException)) {
            mex = new MqttException(MqttException.REASON_CODE_CONNECTION_LOST, ex);
        } else {
            mex = (MqttException) ex;
        }

        mIsRunning = false;
        mClientComms.shutdownConnection(null, mex);
    }
}
