/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import androidx.annotation.NonNull;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.MqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttToken;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttPublish;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttWireMessage;

import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Provides a "token" based system for storing and tracking actions across
 * multiple threads.
 * When a message is sent, a token is associated with the message
 * and saved using the {@link #saveToken(MqttToken, MqttWireMessage)} method. Anyone interested
 * in tacking the state can call one of the wait methods on the token or using
 * the asynchronous listener callback method on the operation.
 * The {@link CommsReceiver} class, on another thread, reads responses back from
 * the network. It uses the response to find the relevant token, which it can then
 * notify.
 * <p/>
 * Note:
 * Ping, connect and disconnect do not have a unique message id as
 * only one outstanding request of each type is allowed to be outstanding
 * <p/>
 * ClientComms가 소유.
 * 각 Op는 thread-safe.
 */
public class CommsTokenStore {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "CommsTokenStore";

    /**
     * Maps message-specific data (usually message IDs) to tokens
     * 맵.
     */
    private final Hashtable<String, MqttToken> mTokenMap;
    /**
     * save 시도 시 발생시킬 예외.
     * 연결 종료 후 설정.
     */
    private MqttException mClosedResponse = null;

    public CommsTokenStore() {
        this.mTokenMap = new Hashtable<String, MqttToken>();
    }

    /**
     * Based on the message type that has just been received return the associated
     * token from the token store or null if one does not exist.
     *
     * @param message whose token is to be returned
     * @return token for the requested message
     */
    public MqttToken getToken(MqttWireMessage message) {
        String key = message.getKey();
        return mTokenMap.get(key);
    }

    public MqttToken getToken(String key) {
        return mTokenMap.get(key);
    }

    public MqttToken removeToken(MqttWireMessage message) {
        if (message != null) {
            return removeToken(message.getKey());
        }
        return null;
    }

    public MqttToken removeToken(String key) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[removeToken] " + MessageFormat.format("key={0}", new Object[]{key}));
        }

        if (key != null) {
            synchronized (mTokenMap) {
                MqttToken tok = mTokenMap.get(key);
                if (tok != null) {
                    synchronized (tok) {

                        return mTokenMap.remove(key);
                    }
                }
            }
        }
        return null;
    }

    /**
     * Restores a token after a client restart. This method could be called
     * for a SEND of CONFIRM, but either way, the original SEND is what's
     * needed to re-build the token.
     */
    protected MqttDeliveryToken restoreToken(MqttPublish message) {
        MqttDeliveryToken token;
        synchronized (mTokenMap) {
            String key = Integer.toString(message.getMessageId());
            if (this.mTokenMap.containsKey(key)) {
                token = (MqttDeliveryToken) this.mTokenMap.get(key);
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[restoreToken] " + MessageFormat.format("existing key={0} message={1} token={2}", new Object[]{key, message, token}));
                }
            } else {
                token = new MqttDeliveryToken();
                token.internalTok.setKey(key);
                this.mTokenMap.put(key, token);
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[restoreToken] " + MessageFormat.format("creating new token key={0} message={1} token={2}", new Object[]{key, message, token}));
                }
            }
        }
        return token;
    }

    // For outbound messages store the token in the token store 
    // For pubrel use the existing publish token 
    protected void saveToken(MqttToken token, MqttWireMessage message) throws MqttException {

        synchronized (mTokenMap) {
            if (mClosedResponse == null) {
                String key = message.getKey();
                if (DEBUG) {
                    LogEx.v(LOG_TAG, "[saveToken] " + MessageFormat.format("key={0} message={1}", new Object[]{key, message}));
                }

                saveToken(token, key);
            } else {
                throw mClosedResponse;
            }
        }
    }

    protected void saveToken(MqttToken token, String key) {

        synchronized (mTokenMap) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[saveToken] " + MessageFormat.format("key={0} token={1}", new Object[]{key, token.toString()}));
            }
            token.internalTok.setKey(key);
            mTokenMap.put(key, token);
        }
    }

    protected void quiesce(MqttException quiesceResponse) {

        synchronized (mTokenMap) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[quiesce] " + MessageFormat.format("resp={0}", new Object[]{quiesceResponse}));
            }

            mClosedResponse = quiesceResponse;
        }
    }

    public void open() {

        synchronized (mTokenMap) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[open] " + ">");
            }

            mClosedResponse = null;
        }
    }

    public MqttDeliveryToken[] getOutstandingDelTokens() {

        synchronized (mTokenMap) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[getOutstandingDelTokens] " + ">");
            }

            Vector<MqttToken> list = new Vector<MqttToken>();
            Enumeration<MqttToken> enumeration = mTokenMap.elements();
            MqttToken token;
            while (enumeration.hasMoreElements()) {
                token = enumeration.nextElement();
                if (token != null
                        && token instanceof MqttDeliveryToken
                        && !token.internalTok.isNotified()) {

                    list.addElement(token);
                }
            }

            MqttDeliveryToken[] result = new MqttDeliveryToken[list.size()];
            return list.toArray(result);
        }
    }

    /**
     * TokenMap 내 모두를 리턴.
     *
     * @return
     */
    @NonNull
    public Vector<MqttToken> getOutstandingTokens() {

        synchronized (mTokenMap) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[getOutstandingTokens] " + ">");
            }

            Vector<MqttToken> list = new Vector<MqttToken>();
            Enumeration<MqttToken> enumeration = mTokenMap.elements();
            MqttToken token;
            while (enumeration.hasMoreElements()) {
                token = enumeration.nextElement();
                if (token != null) {
                    list.addElement(token);
                }
            }
            return list;
        }
    }

    /**
     * Empties the token store without notifying any of the tokens.
     */
    public void clear() {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[clear] " + MessageFormat.format("> {0} tokens", new Object[]{Integer.valueOf(mTokenMap.size())}));
        }
        synchronized (mTokenMap) {
            mTokenMap.clear();
        }
    }

    public int count() {
        synchronized (mTokenMap) {
            return mTokenMap.size();
        }
    }

    @NonNull
    @Override
    public String toString() {
        String lineSep = System.getProperty("line.separator", "\n");
        StringBuffer toks = new StringBuffer();
        synchronized (mTokenMap) {
            Enumeration<MqttToken> enumeration = mTokenMap.elements();
            MqttToken token;
            while (enumeration.hasMoreElements()) {
                token = enumeration.nextElement();
                toks.append("{" + token.internalTok + "}" + lineSep);
            }
            return toks.toString();
        }
    }
}
