/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import android.net.Network;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.MqttException;

import java.io.IOException;
import java.text.MessageFormat;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocket;

/**
 * A network module for connecting over SSL.
 */
public class SSLNetworkModule extends TCPNetworkModule {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "SSLNetworkModule";

    private String[] mEnabledCiphers;
    private int handshakeTimeoutSecs;

    /**
     * Constructs a new SSLNetworkModule using the specified host and
     * port. The supplied SSLSocketFactory is used to supply the network
     * socket.
     */
    public SSLNetworkModule(SocketFactory factory, String host, int port, String resourceContext, Network network) {
        super(factory, host, port, resourceContext, network);
    }

    /**
     * Returns the enabled cipher suites.
     */
    public String[] getEnabledCiphers() {
        return mEnabledCiphers;
    }

    /**
     * Sets the enabled cipher suites on the underlying network socket.
     */
    public void setEnabledCiphers(String[] enabledCiphers) {
        mEnabledCiphers = enabledCiphers;
        if (mSocket != null && enabledCiphers != null) {
            if (DEBUG) {
                StringBuilder builder = new StringBuilder();
                for (int i = 0; i < enabledCiphers.length; i++) {
                    if (i > 0) {
                        builder.append(",");
                    }
                    builder.append(enabledCiphers[i]);
                }
                LogEx.v(LOG_TAG, "[setEnabledCiphers] " + MessageFormat.format("setEnabledCiphers ciphers={0}", new Object[]{builder.toString()}));
            }
            ((SSLSocket) mSocket).setEnabledCipherSuites(enabledCiphers);
        }
    }

    public void setSSLhandshakeTimeout(int timeout) {
        super.setConnectTimeout(timeout);
        this.handshakeTimeoutSecs = timeout;
    }

    @Override
    public void start() throws IOException, MqttException {
        super.start();
        setEnabledCiphers(mEnabledCiphers);
        int soTimeout = mSocket.getSoTimeout();
        if (soTimeout == 0) {
            // RTC 765: Set a timeout to avoid the SSL handshake being blocked indefinitely
            mSocket.setSoTimeout(this.handshakeTimeoutSecs * 1000);
        }
        ((SSLSocket) mSocket).startHandshake();
        // reset timeout to default value
        mSocket.setSoTimeout(soTimeout);
    }
}
