/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * <p>
 * Contributors:
 * Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal;

import android.net.Network;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.comm.lib.utils.LogExUtils;
import com.skt.prod.dialer.business.ProdReleaseManager;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.MqttException;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.text.MessageFormat;

import javax.net.SocketFactory;

/**
 * A network module for connecting over TCP.
 */
public class TCPNetworkModule implements NetworkModule {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "TCPNetworkModule";

    protected Socket mSocket;
    private SocketFactory factory;
    private String host;
    private int port;
    private int conTimeout;
    private Network mNetwork;

    /**
     * Constructs a new TCPNetworkModule using the specified host and
     * port. The supplied SocketFactory is used to supply the network
     * socket.
     */
    public TCPNetworkModule(SocketFactory factory, String host, int port, String resourceContext, Network network) {
        this.factory = factory;
        this.host = host;
        this.port = port;
        this.mNetwork = network;
    }

    /**
     * Starts the module, by creating a TCP socket to the server.
     * 소켓 만들고 연결.
     */
    @Override
    public void start() throws IOException, MqttException {
        try {
            if (LogExUtils.isLogExInfo()) {
                // @TRACE 252=connect to host {0} port {1} timeout {2}
                LogEx.i(LOG_TAG, "[start] " + MessageFormat.format("connect to host {0} port {1} timeout {2}", new Object[]{host, Integer.valueOf(port), Long.valueOf(conTimeout * 1000)}));
            }
            SocketAddress sockaddr;
            mSocket = factory.createSocket();
            // IMS 로 이용!!
            if (ProdReleaseManager.useImsApn()
                    && mNetwork != null) {
                if (LogExUtils.isLogExInfo()) {
                    LogEx.i(LOG_TAG, "[start] USE IMS");
                }
                InetAddress inetAddress = mNetwork.getByName(host);
                if (inetAddress != null && inetAddress.isLoopbackAddress() == false) {
                    sockaddr = new InetSocketAddress(inetAddress, port);
                    if (LogExUtils.isLogExInfo()) {
                        LogEx.i(LOG_TAG, "[start] IMS DNS query");
                    }
                } else {
                    sockaddr = new InetSocketAddress(host, port);
                }
                mNetwork.bindSocket(mSocket);
            } else {
                sockaddr = new InetSocketAddress(host, port);
            }
            mSocket.connect(sockaddr, conTimeout * 1000);

            // SetTcpNoDelay was originally set ot true disabling Nagle's algorithm. 
            // This should not be required.
//            socket.setTcpNoDelay(true);    // TCP_NODELAY on, which means we do not use Nagle's algorithm
        } catch (ConnectException ex) {
            if (LogExUtils.isLogExError()) {
                LogEx.e(LOG_TAG, "[start] " + "Failed to create TCP socket");
            }
            throw new MqttException(MqttException.REASON_CODE_SERVER_CONNECT_ERROR, ex);
        }
    }

    @Override
    public InputStream getInputStream() throws IOException {
        return mSocket.getInputStream();
    }

    @Override
    public OutputStream getOutputStream() throws IOException {
        return mSocket.getOutputStream();
    }

    /**
     * Stops the module, by closing the TCP socket.
     * 소켓 닫기.
     */
    @Override
    public void stop() throws IOException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[stop] " + MessageFormat.format("connect to host {0} port {1}", new Object[]{host, Integer.valueOf(port)}));
        }
        if (mSocket != null) {
            mSocket.close();
        }
    }

    /**
     * Set the maximum time to wait for a socket to be established
     */
    public void setConnectTimeout(int timeout) {
        this.conTimeout = timeout;
    }
}
