/*******************************************************************************
 * Copyright (c) 2014 IBM Corp.
 * <p>
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * <p>
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 */

package org.eclipse.paho.client.mqttv3.internal;

import androidx.annotation.NonNull;

import com.skt.prod.comm.lib.utils.LogEx;
import com.skt.prod.phone.lib.business.AbstractReleaseManager;

import org.eclipse.paho.client.mqttv3.IMqttActionListener;
import org.eclipse.paho.client.mqttv3.IMqttAsyncClient;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttAck;
import org.eclipse.paho.client.mqttv3.internal.wire.MqttWireMessage;

import java.text.MessageFormat;

public class Token {
    private static final boolean DEBUG = AbstractReleaseManager.DEBUG;
    private static String LOG_TAG = "Token";

    volatile private boolean mCompleted = false;
    private boolean mPendingComplete = false;
    private boolean sent = false;

    private final Object mResponseLock = new Object();
    private final Object mSentLock = new Object();

    protected MqttMessage mMessage = null;
    private MqttWireMessage mResponseMessage = null;
    private MqttException mException = null;
    private String[] topics = null;

    private String key;

    private IMqttAsyncClient mClient = null;
    private IMqttActionListener callback = null;

    private Object mUserContext = null;

    public int mMessageID = 0;
    public boolean mIsNotified = false;

    public Token() {
    }

    public int getMessageID() {
        return mMessageID;
    }

    public void setMessageID(int messageID) {
        this.mMessageID = messageID;
    }

    public boolean checkResult() throws MqttException {
        if (getException() != null) {
            throw getException();
        }
        return true;
    }

    public MqttException getException() {
        return mException;
    }

    public boolean isComplete() {
        return mCompleted;
    }

    protected boolean isCompletePending() {
        return mPendingComplete;
    }

    protected boolean isInUse() {
        return (getClient() != null && !isComplete());
    }

    public void setActionCallback(IMqttActionListener listener) {
        this.callback = listener;

    }

    public IMqttActionListener getActionCallback() {
        return callback;
    }

    public void waitForCompletion(long timeout) throws MqttException {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[waitForCompletion] " + MessageFormat.format("key={0} wait max={1} token={2}", new Object[]{getKey(), Long.valueOf(timeout), this}));
        }

        MqttWireMessage resp = waitForResponse(timeout);
        if (resp == null && !mCompleted) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[waitForCompletion] " + MessageFormat.format("key={0} timed out token={1}", new Object[]{getKey(), this}));
            }
            mException = new MqttException(MqttException.REASON_CODE_CLIENT_TIMEOUT);
            throw mException;
        }
        checkResult();
    }

    /**
     * Waits for the message delivery to complete, but doesn't throw an exception
     * in the case of a NACK.  It does still throw an exception if something else
     * goes wrong (e.g. an IOException).  This is used for packets like CONNECT,
     * which have useful information in the ACK that needs to be accessed.
     */
    protected MqttWireMessage waitForResponse(long timeout) throws MqttException {
        synchronized (mResponseLock) {
            if (DEBUG) {
                LogEx.v(LOG_TAG, "[waitForResponse] " + MessageFormat.format(">key={0} timeout={1} sent={2} completed={3} hasException={4} response={5} token={6}",
                        new Object[]{Long.valueOf(timeout), Boolean.valueOf(sent), Boolean.valueOf(mCompleted), (mException == null) ? "false" : "true", mResponseMessage, this}));
            }

            while (!this.mCompleted) {
                if (this.mException == null) {
                    try {
                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[waitForResponse] " + MessageFormat.format("key={0} wait max={1}",
                                    new Object[]{getKey(), Long.valueOf(timeout)}));
                        }

                        if (timeout <= 0) {
                            mResponseLock.wait(); // 무한정 대기.
                        } else {
                            mResponseLock.wait(timeout);
                        }
                    } catch (InterruptedException e) {
                        mException = new MqttException(e);
                    }
                }
                if (!this.mCompleted) {
                    if (this.mException != null) {
                        if (DEBUG) {
                            LogEx.v(LOG_TAG, "[waitForResponse] " + "failed with exception");
                        }
                        throw mException;
                    }

                    if (timeout > 0) {
                        // time up and still not completed
                        break;
                    }
                }
            }
        }
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[waitForResponse] " + MessageFormat.format("key={0} response={1}",
                    new Object[]{getKey(), this.mResponseMessage}));
        }
        return this.mResponseMessage;
    }

    /**
     * Mark the token as complete and ready for users to be notified.
     *
     * @param msg response message. Optional - there are no response messages for some flows
     * @param ex  if there was a problem store the exception in the token.
     */
    protected void markComplete(MqttWireMessage msg, MqttException ex) {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[markComplete] " + MessageFormat.format(">key={0} response={1} excep={2}",
                    new Object[]{getKey(), msg, ex}));
        }

        synchronized (mResponseLock) {
            // ACK means that everything was OK, so mark the message for garbage collection.
            if (msg instanceof MqttAck) {
                this.mMessage = null;
            }
            this.mPendingComplete = true;
            this.mResponseMessage = msg;
            this.mException = ex;
        }
    }

    /**
     * Notifies this token that a response message (an ACK or NACK) has been
     * received.
     */
    protected void notifyComplete() {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[notifyComplete] " + MessageFormat.format(">key={0} response={1} excep={2}",
                    new Object[]{getKey(), this.mResponseMessage, this.mException}));
        }

        synchronized (mResponseLock) {
            // If pending complete is set then normally the token can be marked
            // as complete and users notified. An abnormal error may have
            // caused the client to shutdown between pending complete being set
            // and notifying the user.  In this case - the action must be failed.
            if (mException == null && mPendingComplete) {
                mCompleted = true;
                mPendingComplete = false;
            } else {
                mPendingComplete = false;
            }

            mResponseLock.notifyAll();
        }
        synchronized (mSentLock) {
            sent = true;
            mSentLock.notifyAll();
        }
    }

    public void waitUntilSent() throws MqttException {
        synchronized (mSentLock) {
            synchronized (mResponseLock) {
                if (this.mException != null) {
                    throw this.mException;
                }
            }
            if (!sent) {
                try {
                    if (DEBUG) {
                        LogEx.v(LOG_TAG, "[waitUntilSent] " + MessageFormat.format("wait key={0}",
                                new Object[]{getKey()}));
                    }

                    mSentLock.wait();
                } catch (InterruptedException ignored) {
                }
            }

            if (!sent) {
                if (this.mException == null) {
                    throw ExceptionHelper.createMqttException(MqttException.REASON_CODE_UNEXPECTED_ERROR);
                }
                throw this.mException;
            }
        }
    }

    /**
     * Notifies this token that the associated message has been sent
     * (i.e. written to the TCP/IP socket).
     */
    protected void notifySent() {
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[notifySent] " + MessageFormat.format("> key={0}",
                    new Object[]{getKey()}));
        }
        synchronized (mResponseLock) {
            this.mResponseMessage = null;
            this.mCompleted = false;
        }
        synchronized (mSentLock) {
            sent = true;
            mSentLock.notifyAll();
        }
    }

    public IMqttAsyncClient getClient() {
        return mClient;
    }

    protected void setClient(IMqttAsyncClient client) {
        this.mClient = client;
    }

    public void reset() throws MqttException {
        if (isInUse()) {
            // Token is already in use - cannot reset
            throw new MqttException(MqttException.REASON_CODE_TOKEN_INUSE);
        }
        if (DEBUG) {
            LogEx.v(LOG_TAG, "[reset] " + MessageFormat.format("> key={0}",
                    new Object[]{getKey()}));
        }

        mClient = null;
        mCompleted = false;
        mResponseMessage = null;
        sent = false;
        mException = null;
        mUserContext = null;
    }

    public MqttMessage getMessage() {
        return mMessage;
    }

    public MqttWireMessage getWireMessage() {
        return mResponseMessage;
    }


    public void setMessage(MqttMessage msg) {
        this.mMessage = msg;
    }

    public String[] getTopics() {
        return topics;
    }

    public void setTopics(String[] topics) {
        this.topics = topics;
    }

    public Object getUserContext() {
        return mUserContext;
    }

    public void setUserContext(Object userContext) {
        this.mUserContext = userContext;
    }

    public void setKey(String key) {
        this.key = key;
    }

    public String getKey() {
        return key;
    }

    public void setException(MqttException exception) {
        synchronized (mResponseLock) {
            this.mException = exception;
        }
    }

    public boolean isNotified() {
        return mIsNotified;
    }

    public void setNotified(boolean notified) {
        this.mIsNotified = notified;
    }

    @NonNull
    public String toString() {
        StringBuffer tok = new StringBuffer();
        tok.append("key=" + getKey());
        tok.append(" ,topics=");
        if (getTopics() != null) {
            for (int i = 0; i < getTopics().length; i++) {
                tok.append(getTopics()[i] + ", ");
            }
        }
        tok.append(" ,usercontext=" + getUserContext());
        tok.append(" ,isComplete=" + isComplete());
        tok.append(" ,isNotified=" + isNotified());
        tok.append(" ,exception=" + getException());
        tok.append(" ,actioncallback=" + getActionCallback());

        return tok.toString();
    }

}
