/*******************************************************************************
 * Copyright (c) 2009, 2014 IBM Corp.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution. 
 *
 * The Eclipse Public License is available at 
 *    http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at 
 *   http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * Contributors:
 *    Dave Locke - initial API and implementation and/or initial documentation
 */
package org.eclipse.paho.client.mqttv3.internal.wire;

import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * An on-the-wire representation of an MQTT SEND message.
 */
public class MqttPublish extends MqttPersistableWireMessage {

    private MqttMessage mMessage;
    private String mTopicName;

    private byte[] encodedPayload = null;

    public MqttPublish(String name, MqttMessage message) {
        super(MqttWireMessage.MESSAGE_TYPE_PUBLISH);
        mTopicName = name;
        this.mMessage = message;
    }

    /**
     * Constructs a new MqttPublish object.
     *
     * @param info the message info byte
     * @param data the variable header and payload bytes
     */
    public MqttPublish(byte info, byte[] data) throws MqttException, IOException {
        super(MqttWireMessage.MESSAGE_TYPE_PUBLISH);
        mMessage = new MqttReceivedMessage();
        // Qos
        mMessage.setQos((info >> 1) & 0x03);
        // Retain
        if ((info & 0x01) == 0x01) {
            mMessage.setRetained(true);
        }
        // dup
        if ((info & 0x08) == 0x08) {
            ((MqttReceivedMessage) mMessage).setDuplicate(true);
        }

        ByteArrayInputStream bais = new ByteArrayInputStream(data);
        CountingInputStream counter = new CountingInputStream(bais);
        DataInputStream dis = new DataInputStream(counter);
        mTopicName = decodeUTF8(dis);
        if (mMessage.getQos() > 0) {
            mMsgId = dis.readUnsignedShort();
        }
        byte[] payload = new byte[data.length - counter.getCounter()];
        dis.readFully(payload);
        dis.close();
        mMessage.setPayload(payload);
    }

    public String toString() {

        // Convert the first few bytes of the payload into a hex string
        StringBuffer hex = new StringBuffer();
        byte[] payload = mMessage.getPayload();
        int limit = Math.min(payload.length, 100);
        for (int i = 0; i < limit; i++) {
            byte b = payload[i];
            String ch = Integer.toHexString(b);
            if (ch.length() == 1) {
                ch = "0" + ch;
            }
            hex.append(ch);
        }

        // It will not always be possible to convert the binary payload into
        // characters, but never-the-less we attempt to do this as it is often
        // useful
        String string = null;
        try {
            string = new String(payload, 0, limit, "UTF-8");
        } catch (Exception e) {
            string = "?";
        }

        StringBuffer sb = new StringBuffer();
        sb.append(super.toString());
        sb.append(" qos:" + mMessage.getQos());
        if (mMessage.getQos() > 0) {
            sb.append(" mMsgId:" + mMsgId);
        }
        sb.append(" retained:" + mMessage.isRetained());
        sb.append(" dup:" + duplicate);
        sb.append(" topic:\"" + mTopicName + "\"");
        sb.append(" payload:[hex:" + hex);
        sb.append(" utf8:\"" + string + "\"");
        sb.append(" length:" + payload.length + "]");

        return sb.toString();
    }

    protected byte getMessageInfo() {
        byte info = (byte) (mMessage.getQos() << 1);
        if (mMessage.isRetained()) {
            info |= 0x01;
        }
        if (mMessage.isDuplicate() || duplicate) {
            info |= 0x08;
        }

        return info;
    }

    public String getTopicName() {
        return mTopicName;
    }

    public MqttMessage getMessage() {
        return mMessage;
    }

    protected static byte[] encodePayload(MqttMessage message) {
        return message.getPayload();
    }

    public byte[] getPayload() throws MqttException {
        if (encodedPayload == null) {
            encodedPayload = encodePayload(mMessage);
        }
        return encodedPayload;
    }

    public int getPayloadLength() {
        int length = 0;
        try {
            length = getPayload().length;
        } catch (MqttException ignored) {
        }
        return length;
    }

    public void setMessageId(int msgId) {
        super.setMessageId(msgId);
        if (mMessage instanceof MqttReceivedMessage) {
            ((MqttReceivedMessage) mMessage).setMessageId(msgId);
        }
    }

    protected byte[] getVariableHeader() throws MqttException {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            DataOutputStream dos = new DataOutputStream(baos);
            encodeUTF8(dos, mTopicName);
            if (mMessage.getQos() > 0) {
                dos.writeShort(mMsgId);
            }
            dos.flush();
            return baos.toByteArray();
        } catch (IOException ex) {
            throw new MqttException(ex);
        }
    }

    public boolean isMessageIdRequired() {
        // all publishes require a message ID as it's used as the key to the token store
        return true;
    }
}